"""Refiner is the refinement module public interface. RefinerFactory is
what should usually be used to construct a Refiner."""


from __future__ import annotations

import copy
import logging
import math

import psutil

import libtbx
from dxtbx.model.experiment_list import ExperimentList
from libtbx.phil import parse

import dials.util
from dials.algorithms.refinement import DialsRefineConfigError
from dials.algorithms.refinement.constraints import ConstraintManagerFactory
from dials.algorithms.refinement.engine import AdaptLstbx, refinery_phil_str
from dials.algorithms.refinement.parameterisation import (
    build_prediction_parameterisation,
)
from dials.algorithms.refinement.parameterisation import (
    phil_str as parameterisation_phil_str,
)
from dials.algorithms.refinement.parameterisation.autoreduce import AutoReduce
from dials.algorithms.refinement.parameterisation.parameter_report import (
    ParameterReporter,
)
from dials.algorithms.refinement.prediction.managed_predictors import (
    ExperimentsPredictorFactory,
)
from dials.algorithms.refinement.refinement_helpers import ordinal_number, string_sel
from dials.algorithms.refinement.reflection_manager import ReflectionManagerFactory
from dials.algorithms.refinement.reflection_manager import (
    phil_str as reflections_phil_str,
)
from dials.algorithms.refinement.restraints import RestraintsParameterisation
from dials.algorithms.refinement.target import TargetFactory
from dials.algorithms.refinement.target import phil_str as target_phil_str
from dials.array_family import flex

logger = logging.getLogger(__name__)

# The include scope directive does not work here. For example:
#
#   include scope dials.algorithms.refinement.outlier_detection.phil_scope
#
# results in:
#
#   AttributeError: 'module' object has no attribute 'refinement'
#
# to work around this, just include external phil scopes as strings
format_data = {
    "reflections_phil": reflections_phil_str,
    "target_phil": target_phil_str,
    "parameterisation_phil": parameterisation_phil_str,
    "refinery_phil": refinery_phil_str,
}
phil_scope = parse(
    """
refinement
  .help = "Parameters to configure the refinement"
{

  mp
    .expert_level = 2
  {
    nproc = 1
      .type = int(value_min=1)
      .help = "The number of processes to use. Not all choices of refinement"
              "engine support nproc > 1. Where multiprocessing is possible,"
              "it is helpful only in certain circumstances, so this is not"
              "recommended for typical use."
  }

  parameterisation
    .help = "Parameters to control the parameterisation of experimental models"
  {
    %(parameterisation_phil)s
  }

  %(refinery_phil)s

  target
    .help = "Parameters to configure the target function"
    .expert_level = 1
  {
    %(target_phil)s
  }

  reflections
    .help = "Parameters used by the reflection manager"
  {
    %(reflections_phil)s
  }

}
"""
    % format_data,
    process_includes=True,
)

RAD2DEG = 180 / math.pi


def _copy_experiments_for_refining(experiments):
    """
    Make a partial copy of experiments, copying properties used in refinement.

    Any experiment property that can be altered by refinement e.g. Beam,
    Detector and Goniometer will be deep-copied, whereas anything that the
    refiner doesn't touch (e.g. Scan, ImageSet) will be left as original
    references.

    This makes it safe to pass an object into the refiner or get an object
    out of the refiner without having to worry about your copy being
    unexpectedly altered, but saves time by avoiding the copying of potentially
    expensive experiment properties (e.g. ImageSet and its attributes).

    Args
      experiments (Experiment or ExperimentList or Iterable[Experiment]):

    Returns:
      ExperimentList: The copied experiments in new ExperimentList
    """
    # Look for a non-list e.g. a single experiment and convert to a list
    if not hasattr(experiments, "__iter__"):
        experiments = [experiments]
    out_list = ExperimentList()
    # Save a map of object id to copies so shared objects remain shared
    id_memo = {}

    # Copy each experiment individually
    for experiment in experiments:
        # Be inclusive about the initial copy
        new_exp = copy.copy(experiment)

        # Ensure every 'refined' attribute is uniquely copied
        for model in ["beam", "goniometer", "detector", "crystal"]:
            original = getattr(experiment, model)
            if id(original) not in id_memo:
                id_memo[id(original)] = copy.deepcopy(original)
            # assign the new copy to the experiment
            setattr(new_exp, model, id_memo[id(original)])

        # Collect this together
        out_list.append(new_exp)

    return out_list


def _trim_scans_to_observations(experiments, reflections):
    """Check the range of each scan matches the range of observed data and
    trim the scan to match if it is too wide"""

    # Get observed image number (or at least observed phi)
    obs_phi = reflections["xyzobs.mm.value"].parts()[2]
    try:
        obs_z = reflections["xyzobs.px.value"].parts()[2]
    except KeyError:
        obs_z = None

    # Get z_min and z_max from shoeboxes if present
    try:
        shoebox = reflections["shoebox"]
        bb = shoebox.bounding_boxes()
        z_min, z_max = bb.parts()[4:]
        if z_min.all_eq(0):
            shoebox = None
    except KeyError:
        shoebox = None

    for iexp, exp in enumerate(experiments):

        sel = reflections["id"] == iexp
        isel = sel.iselection()
        if obs_z is not None:
            exp_z = obs_z.select(isel)
        else:
            exp_phi = obs_phi.select(isel)
            exp_z = exp.scan.get_array_index_from_angle(exp_phi, deg=False)

        start, stop = exp.scan.get_array_range()
        min_exp_z = flex.min(exp_z)
        max_exp_z = flex.max(exp_z)

        # If observed array range is correct, skip to next experiment
        if int(min_exp_z) == start and int(math.ceil(max_exp_z)) == stop:
            continue

        # Extend array range either by shoebox size, or 0.5 deg if shoebox not available
        if shoebox is not None:
            obs_start = flex.min(z_min.select(isel))
            obs_stop = flex.max(z_max.select(isel))
        else:
            obs_start = int(min_exp_z)
            obs_stop = int(math.ceil(max_exp_z))
            half_deg_in_images = int(math.ceil(0.5 / exp.scan.get_oscillation()[1]))
            obs_start -= half_deg_in_images
            obs_stop += half_deg_in_images

        # Convert obs_start, obs_stop from position in array range to integer image number
        if obs_start > start or obs_stop < stop:
            im_start = max(start, obs_start) + 1
            im_stop = min(obs_stop, stop)

            logger.warning(
                "The reflections for experiment {0} do not fill the scan range. The scan will be trimmed "
                "to images {{{1},{2}}} to match the range of observed data".format(
                    iexp, im_start, im_stop
                )
            )

            # Ensure the scan is unique to this experiment and set trimmed limits
            exp.scan = copy.deepcopy(exp.scan)
            new_oscillation = (
                exp.scan.get_angle_from_image_index(im_start),
                exp.scan.get_oscillation()[1],
            )
            exp.scan.set_image_range((im_start, im_stop))
            exp.scan.set_oscillation(new_oscillation)

    return experiments


class RefinerFactory:
    """Factory class to create refiners"""

    @staticmethod
    def _filter_reflections(reflections):
        """Return a copy of the input reflections filtered to keep only
        those columns that are required by refinement"""

        cols = [
            "id",
            "miller_index",
            "panel",
            "s1",
            "xyzobs.mm.value",
            "xyzobs.px.value",
            "xyzcal.px",
            "xyzobs.mm.variance",
            "flags",
            "shoebox",
            "delpsical.weights",
        ]
        # NB xyzobs.px.value & xyzcal.px required by SauterPoon outlier rejector
        # NB delpsical.weights is used by ExternalDelPsiWeightingStrategy
        rt = flex.reflection_table()

        # copy columns to the new table. Could use the select method
        # for this except that 's1' is optional in the input so would want
        # to copy that in like this if present anyway
        for k in cols:
            if k in reflections:
                rt[k] = reflections[k]

        return rt

    @classmethod
    def from_parameters_data_experiments(cls, params, reflections, experiments):
        # TODO Checks on the input
        # E.g. does every experiment contain at least one overlapping model with at
        # least one other experiment? Are all the experiments either rotation series
        # or stills (the combination of both not yet supported)?

        # copy the experiments
        experiments = _copy_experiments_for_refining(experiments)

        # copy and filter the reflections
        reflections = cls._filter_reflections(reflections)

        return cls._build_components(params, reflections, experiments)

    @classmethod
    def _build_components(cls, params, reflections, experiments):
        """low level build"""

        # Currently a refinement job can only have one parameterisation of the
        # prediction equation. This can either be of the XYDelPsi (stills) type, the
        # XYPhi (scans) type or the scan-varying XYPhi type with a varying crystal
        # model
        single_as_still = params.refinement.parameterisation.treat_single_image_as_still
        exps_are_stills = []
        for exp in experiments:
            if exp.scan is None:
                exps_are_stills.append(True)
            elif exp.scan.get_num_images() == 1:
                if single_as_still:
                    exps_are_stills.append(True)
                elif exp.scan.is_still():
                    exps_are_stills.append(True)
                else:
                    exps_are_stills.append(False)
            else:
                if exp.scan.is_still():
                    raise DialsRefineConfigError("Cannot refine a zero-width scan")
                exps_are_stills.append(False)

        # check experiment types are consistent
        if not all(exps_are_stills[0] == e for e in exps_are_stills):
            raise DialsRefineConfigError("Cannot refine a mixture of stills and scans")
        do_stills = exps_are_stills[0]

        # If experiments are stills, ensure scan-varying refinement won't be attempted
        if do_stills:
            params.refinement.parameterisation.scan_varying = False

        # Refiner does not accept scan_varying=Auto. This is a special case for
        # doing macrocycles of refinement in dials.refine.
        if params.refinement.parameterisation.scan_varying is libtbx.Auto:
            params.refinement.parameterisation.scan_varying = False

        # Calculate reflection block_width for scan-varying refinement. Trim scans
        # to the extent of the observations, if requested.
        if params.refinement.parameterisation.scan_varying:
            if params.refinement.parameterisation.trim_scan_to_observations:
                experiments = _trim_scans_to_observations(experiments, reflections)

            from dials.algorithms.refinement.reflection_manager import BlockCalculator

            block_calculator = BlockCalculator(experiments, reflections)
            if params.refinement.parameterisation.compose_model_per == "block":
                reflections = block_calculator.per_width(
                    params.refinement.parameterisation.block_width, deg=True
                )
            elif params.refinement.parameterisation.compose_model_per == "image":
                reflections = block_calculator.per_image()

        logger.debug("\nBuilding reflection manager")
        logger.debug("Input reflection list size = %d observations", len(reflections))

        # create reflection manager
        refman = ReflectionManagerFactory.from_parameters_reflections_experiments(
            params.refinement.reflections, reflections, experiments, do_stills
        )

        logger.debug(
            "Number of observations that pass initial inclusion criteria = %d",
            refman.get_accepted_refs_size(),
        )
        sample_size = refman.get_sample_size()
        if sample_size > 0:
            logger.debug("Working set size = %d observations", sample_size)
        logger.debug("Reflection manager built\n")

        # configure use of sparse data types
        params = cls.config_sparse(params, experiments)
        do_sparse = params.refinement.parameterisation.sparse

        # create managed reflection predictor
        ref_predictor = ExperimentsPredictorFactory.from_experiments(
            experiments,
            force_stills=do_stills,
            spherical_relp=params.refinement.parameterisation.spherical_relp_model,
        )

        # Predict for the managed observations, set columns for residuals and set
        # the used_in_refinement flag to the predictions
        obs = refman.get_obs()
        ref_predictor(obs)
        x_obs, y_obs, phi_obs = obs["xyzobs.mm.value"].parts()
        x_calc, y_calc, phi_calc = obs["xyzcal.mm"].parts()
        obs["x_resid"] = x_calc - x_obs
        obs["y_resid"] = y_calc - y_obs
        obs["phi_resid"] = phi_calc - phi_obs

        # determine whether to do basic centroid analysis to automatically
        # determine outlier rejection block
        if params.refinement.reflections.outlier.block_width is libtbx.Auto:
            ca = refman.get_centroid_analyser()
            analysis = ca(calc_average_residuals=False, calc_periodograms=False)
        else:
            analysis = None

        # Now predictions and centroid analysis are available, so we can finalise
        # the reflection manager
        refman.finalise(analysis)

        # Create model parameterisations
        logger.debug("Building prediction equation parameterisation")
        pred_param = build_prediction_parameterisation(
            params.refinement.parameterisation, experiments, refman, do_stills
        )

        # Build a constraints manager, if requested
        cmf = ConstraintManagerFactory(params, pred_param)
        constraints_manager = cmf()

        # Test for parameters that have too little data to refine and act accordingly
        autoreduce = AutoReduce(
            params.refinement.parameterisation.auto_reduction,
            pred_param,
            refman,
            constraints_manager,
            cmf,
        )
        autoreduce()

        # if reduction was done, constraints_manager will have changed
        constraints_manager = autoreduce.constraints_manager

        # Build a restraints parameterisation (if requested).
        # Only unit cell restraints are supported at the moment.
        restraints_parameterisation = cls.config_restraints(
            params.refinement.parameterisation, pred_param
        )

        # Parameter reporting
        logger.debug("Prediction equation parameterisation built")
        logger.debug("Parameter order : name mapping")
        for i, e in enumerate(pred_param.get_param_names()):
            logger.debug("Parameter %03d : %s", i + 1, e)

        param_reporter = ParameterReporter(
            pred_param.get_detector_parameterisations(),
            pred_param.get_beam_parameterisations(),
            pred_param.get_crystal_orientation_parameterisations(),
            pred_param.get_crystal_unit_cell_parameterisations(),
            pred_param.get_goniometer_parameterisations(),
        )

        # Create target function
        logger.debug("Building target function")
        target = cls.config_target(
            params.refinement.target,
            experiments,
            refman,
            ref_predictor,
            pred_param,
            restraints_parameterisation,
            do_stills,
            do_sparse,
        )
        logger.debug("Target function built")

        # create refinery
        logger.debug("Building refinement engine")
        refinery = cls.config_refinery(params, target, pred_param, constraints_manager)
        logger.debug("Refinement engine built")

        nparam = len(pred_param)
        ndim = target.dim
        nref = len(refman.get_matches())
        logger.info(
            "There are %s parameters to refine against %s reflections in %s dimensions",
            nparam,
            nref,
            ndim,
        )

        if not params.refinement.parameterisation.sparse and isinstance(
            refinery, AdaptLstbx
        ):
            dense_jacobian_gigabytes = (
                nparam * nref * ndim * flex.double.element_size()
            ) / 1e9
            avail_memory_gigabytes = psutil.virtual_memory().available / 1e9
            # Report if the Jacobian requires a large amount of storage
            if (
                dense_jacobian_gigabytes > 0.2 * avail_memory_gigabytes
                or dense_jacobian_gigabytes > 0.5
            ):
                logger.info(
                    "Storage of the Jacobian matrix requires %.1f GB",
                    dense_jacobian_gigabytes,
                )

        # build refiner interface and return
        if params.refinement.parameterisation.scan_varying:
            refiner = ScanVaryingRefiner
        else:
            refiner = Refiner
        return refiner(
            experiments, pred_param, param_reporter, refman, target, refinery
        )

    @staticmethod
    def config_sparse(params, experiments):
        """Configure whether to use sparse datatypes"""
        # Automatic selection for sparse parameter
        if params.refinement.parameterisation.sparse == libtbx.Auto:
            if len(experiments) > 1 or params.refinement.parameterisation.scan_varying:
                params.refinement.parameterisation.sparse = True
            else:
                params.refinement.parameterisation.sparse = False
            if params.refinement.refinery.engine == "SparseLevMar":
                params.refinement.parameterisation.sparse = True
            if params.refinement.mp.nproc > 1:
                if params.refinement.refinery.engine != "SparseLevMar":
                    # sparse vectors cannot be pickled, so can't use easy_mp here
                    params.refinement.parameterisation.sparse = False
                else:
                    pass  # but SparseLevMar requires sparse jacobian; does not implement mp
        # Check incompatible selection
        elif (
            params.refinement.parameterisation.sparse and params.refinement.mp.nproc > 1
        ):
            logger.warning(
                "Could not set sparse=True and nproc=%s", params.refinement.mp.nproc
            )
            logger.warning("Resetting sparse=False")
            params.refinement.parameterisation.sparse = False
        return params

    @staticmethod
    def config_restraints(params, pred_param):
        """Given a set of user parameters plus a model parameterisation, create
        restraints plus a parameterisation of these restraints

        Params:
            params: The input PHIL parameters
            pred_param: A PredictionParameters object

        Returns:
            A restraints parameterisation or None
        """

        if not any(
            (
                params.crystal.unit_cell.restraints.tie_to_target,
                params.crystal.unit_cell.restraints.tie_to_group,
            )
        ):
            return None

        if params.scan_varying and not params.crystal.unit_cell.force_static:
            logger.warning("Restraints will be ignored for scan_varying=True")
            return None

        det_params = pred_param.get_detector_parameterisations()
        beam_params = pred_param.get_beam_parameterisations()
        xl_ori_params = pred_param.get_crystal_orientation_parameterisations()
        xl_uc_params = pred_param.get_crystal_unit_cell_parameterisations()
        gon_params = pred_param.get_goniometer_parameterisations()

        rp = RestraintsParameterisation(
            detector_parameterisations=det_params,
            beam_parameterisations=beam_params,
            xl_orientation_parameterisations=xl_ori_params,
            xl_unit_cell_parameterisations=xl_uc_params,
            goniometer_parameterisations=gon_params,
        )

        # Shorten params path
        cell_r = params.crystal.unit_cell.restraints

        for tie in cell_r.tie_to_target:
            if len(tie.values) != 6:
                raise DialsRefineConfigError(
                    "6 cell parameters must be provided as the tie_to_target.values."
                )
            if len(tie.sigmas) != 6:
                raise DialsRefineConfigError(
                    "6 sigmas must be provided as the tie_to_target.sigmas. "
                    "Note that individual sigmas of 0.0 will remove "
                    "the restraint for the corresponding cell parameter."
                )
            if tie.id is None:
                # get one experiment id for each parameterisation to apply to all
                tie.id = [e.get_experiment_ids()[0] for e in xl_uc_params]
            for exp_id in tie.id:
                rp.add_restraints_to_target_xl_unit_cell(exp_id, tie.values, tie.sigmas)

        for tie in cell_r.tie_to_group:
            if len(tie.sigmas) != 6:
                raise DialsRefineConfigError(
                    "6 sigmas must be provided as the tie_to_group.sigmas. "
                    "Note that individual sigmas of 0.0 will remove "
                    "the restraint for the corresponding cell parameter."
                )
            if tie.id is None:
                rp.add_restraints_to_group_xl_unit_cell(tie.target, "all", tie.sigmas)
            else:
                rp.add_restraints_to_group_xl_unit_cell(tie.target, tie.id, tie.sigmas)

        return rp

    @staticmethod
    def config_refinery(params, target, pred_param, constraints_manager):
        """Given a set of parameters, a target class, a prediction
        parameterisation class and a constraints_manager (which could be None),
        build a refinery

        Params:
            params The input parameters

        Returns:
            The refinery instance
        """

        # Shorten parameter path
        options = params.refinement.refinery

        if options.engine == "SimpleLBFGS":
            from dials.algorithms.refinement.engine import SimpleLBFGS as refinery
        elif options.engine == "LBFGScurvs":
            from dials.algorithms.refinement.engine import LBFGScurvs as refinery
        elif options.engine == "GaussNewton":
            from dials.algorithms.refinement.engine import (
                GaussNewtonIterations as refinery,
            )
        elif options.engine == "LevMar":
            from dials.algorithms.refinement.engine import (
                LevenbergMarquardtIterations as refinery,
            )
        elif options.engine == "SparseLevMar":
            from dials.algorithms.refinement.sparse_engine import (
                SparseLevenbergMarquardtIterations as refinery,
            )
        else:
            raise RuntimeError(
                "Refinement engine " + options.engine + " not recognised"
            )

        logger.debug("Selected refinement engine type: %s", options.engine)

        engine = refinery(
            target=target,
            prediction_parameterisation=pred_param,
            constraints_manager=constraints_manager,
            log=options.log,
            tracking=options.journal,
            max_iterations=options.max_iterations,
        )

        if params.refinement.mp.nproc > 1:
            nproc = params.refinement.mp.nproc
            try:
                engine.set_nproc(nproc)
            except NotImplementedError:
                logger.warning(
                    "Could not set nproc=%s for refinement engine of type %s",
                    nproc,
                    options.engine,
                )

        return engine

    # Overload to allow subclasses of RefinerFactory to use a different
    # TargetFactory
    @staticmethod
    def config_target(
        params,
        experiments,
        reflection_manager,
        predictor,
        pred_param,
        restraints_param,
        do_stills,
        do_sparse,
    ):

        target = TargetFactory.from_parameters_and_experiments(
            params,
            experiments,
            reflection_manager,
            predictor,
            pred_param,
            restraints_param,
            do_stills,
            do_sparse,
        )
        return target


class Refiner:
    """Public interface for performing DIALS refinement.

    Public methods:
      run
      rmsds
      get_experiments
      get_matches
      get_param_reporter
      parameter_correlation_plot
      selection_used_for_refinement
      predict_for_reflection_table
      predict_for_indexed

    Notes:
      * The return value of run is a recorded history of the refinement
      * The experiments accessor provides a copy of the experiments used by
        refinement
      * get_matches exposes the function of the same name from the privately
        stored reflection manager
      * The return value of selection_used_for_refinement is a flex.bool
    """

    def __init__(
        self, experiments, pred_param, param_reporter, refman, target, refinery
    ):
        """
        Mandatory arguments:
          experiments - a dxtbx ExperimentList object
          pred_param - An object derived from the PredictionParameterisation class
          param_reporter -A ParameterReporter object
          refman - A ReflectionManager object
          target - An object derived from the Target class
          refinery - An object derived from the Refinery class
        """

        # the experimental models
        self._experiments = experiments

        # refinement module main objects
        self._pred_param = pred_param
        self._refman = refman
        self._target = target
        self._refinery = refinery

        # parameter reporter
        self._param_report = param_reporter

        # Keep track of whether this is stills or scans type refinement
        self.experiment_type = refman.experiment_type

        return

    def get_experiments(self):
        """Return a copy of the current refiner experiments"""
        return _copy_experiments_for_refining(self._experiments)

    def rmsds(self):
        """Return rmsds of the current model"""

        # ensure predictions for the matches are up to date
        self._refinery.prepare_for_step()

        return self._target.rmsds()

    def rmsds_for_reflection_table(self, reflections):
        """Calculate unweighted RMSDs for the specified reflections"""

        # ensure predictions for these reflections are up to date
        preds = self.predict_for_reflection_table(reflections)

        return self._target.rmsds_for_reflection_table(preds)

    def get_matches(self):
        """Delegated to the reflection manager"""

        return self._refman.get_matches()

    def get_free_reflections(self):
        """Delegated to the reflection manager"""

        return self._refman.get_free_reflections()

    def get_param_reporter(self):
        """Get the ParameterReport object linked to this Refiner"""

        return self._param_report

    def get_parameter_correlation_matrix(self, step, col_select=None):
        """Return the correlation matrix between columns of the Jacobian at
        the specified refinement step. The parameter col_select can be used
        to select subsets of the full number of columns. The column labels
        are also returned as a list of strings"""

        corrmats = self._refinery.get_correlation_matrix_for_step(step)
        if corrmats is None:
            return None, None

        all_labels = self._pred_param.get_param_names()

        if col_select is None:
            col_select = list(range(len(all_labels)))
        sel = string_sel(col_select, all_labels)
        labels = [e for e, s in zip(all_labels, sel) if s]
        num_cols = len(labels)
        if num_cols == 0:
            return None, None

        for k, corrmat in corrmats.items():

            assert corrmat.is_square_matrix()

            idx = flex.bool(sel).iselection()
            sub_corrmat = flex.double(flex.grid(num_cols, num_cols))

            for (i, x) in enumerate(idx):
                for (j, y) in enumerate(idx):
                    sub_corrmat[i, j] = corrmat[x, y]

            corrmats[k] = sub_corrmat

        return (corrmats, labels)

    @property
    def history(self):
        """Get the refinement engine's step history"""
        return self._refinery.history

    def print_step_table(self):
        """print useful output about refinement steps in the form of a simple table"""

        logger.info("\nRefinement steps:")

        rmsd_multipliers = []
        header = ["Step", "Nref"]
        for (name, units) in zip(self._target.rmsd_names, self._target.rmsd_units):
            if units == "mm":
                header.append(name + "\n(mm)")
                rmsd_multipliers.append(1.0)
            elif units == "rad":  # convert radians to degrees for reporting
                header.append(name + "\n(deg)")
                rmsd_multipliers.append(RAD2DEG)
            else:  # leave unknown units alone
                header.append(name + "\n(" + units + ")")

        rows = []
        for i in range(self._refinery.history.get_nrows()):
            rmsds = [
                r * m
                for (r, m) in zip(self._refinery.history["rmsd"][i], rmsd_multipliers)
            ]
            rows.append(
                [str(i), str(self._refinery.history["num_reflections"][i])]
                + [f"{r:.5g}" for r in rmsds]
            )

        logger.info(dials.util.tabulate(rows, header))
        logger.info(self._refinery.history.reason_for_termination)

        return

    def print_out_of_sample_rmsd_table(self):
        """print out-of-sample RSMDs per step, if these were tracked"""

        # check if it makes sense to proceed
        if "out_of_sample_rmsd" not in self._refinery.history:
            return
        nref = len(self.get_free_reflections())
        if nref < 10:
            return  # don't do anything if very few refs

        logger.info("\nRMSDs for out-of-sample (free) reflections:")

        rmsd_multipliers = []
        header = ["Step", "Nref"]
        for (name, units) in zip(self._target.rmsd_names, self._target.rmsd_units):
            if units == "mm":
                header.append(name + "\n(mm)")
                rmsd_multipliers.append(1.0)
            elif units == "rad":  # convert radians to degrees for reporting
                header.append(name + "\n(deg)")
                rmsd_multipliers.append(RAD2DEG)
            else:  # leave unknown units alone
                header.append(name + "\n(" + units + ")")

        rows = []
        for i in range(self._refinery.history.get_nrows()):
            rmsds = [
                r * m
                for r, m in zip(
                    self._refinery.history["out_of_sample_rmsd"][i], rmsd_multipliers
                )
            ]
            rows.append([str(i), str(nref)] + [f"{e:.5g}" for e in rmsds])

        logger.info(dials.util.tabulate(rows, header))

        return

    def print_exp_rmsd_table(self):
        """print useful output about refinement steps in the form of a simple table"""

        logger.info("\nRMSDs by experiment:")

        header = ["Exp\nid", "Nref"]
        for (name, units) in zip(self._target.rmsd_names, self._target.rmsd_units):
            if name == "RMSD_X" or name == "RMSD_Y" and units == "mm":
                header.append(name + "\n(px)")
            elif name == "RMSD_Phi" and units == "rad":
                # will convert radians to images for reporting of scans
                header.append("RMSD_Z" + "\n(images)")
            elif units == "rad":
                # will convert other angles in radians to degrees (e.g. for
                # RMSD_DeltaPsi and RMSD_2theta)
                header.append(name + "\n(deg)")
            else:  # skip other/unknown RMSDs
                pass

        rows = []
        for iexp, exp in enumerate(self._experiments):
            detector = exp.detector
            px_sizes = [p.get_pixel_size() for p in detector]
            it = iter(px_sizes)
            px_size = next(it)
            if not all(tst == px_size for tst in it):
                logger.info(
                    "The detector in experiment %d does not have the same pixel "
                    + "sizes on each panel. Skipping...",
                    iexp,
                )
                continue
            px_per_mm = [1.0 / e for e in px_size]

            scan = exp.scan
            try:
                images_per_rad = 1.0 / abs(scan.get_oscillation(deg=False)[1])
            except (AttributeError, ZeroDivisionError):
                images_per_rad = None

            raw_rmsds = self._target.rmsds_for_experiment(iexp)
            if raw_rmsds is None:
                continue  # skip experiments where rmsd cannot be calculated
            num = self._target.get_num_matches_for_experiment(iexp)
            rmsds = []
            for (name, units, rmsd) in zip(
                self._target.rmsd_names, self._target.rmsd_units, raw_rmsds
            ):
                if name == "RMSD_X" and units == "mm":
                    rmsds.append(rmsd * px_per_mm[0])
                elif name == "RMSD_Y" and units == "mm":
                    rmsds.append(rmsd * px_per_mm[1])
                elif name == "RMSD_Phi" and units == "rad":
                    rmsds.append(rmsd * images_per_rad)
                elif units == "rad":
                    rmsds.append(rmsd * RAD2DEG)
            rows.append([str(iexp), str(num)] + [f"{r:.5g}" for r in rmsds])

        if len(rows) > 0:
            logger.info(dials.util.tabulate(rows, header))

        return

    def print_panel_rmsd_table(self):
        """print useful output about refinement steps in the form of a simple table"""

        if len(self._experiments.scans()) > 1:
            logger.warning(
                "Multiple scans present. Only the first scan will be used "
                "to determine the image width for reporting RMSDs"
            )
        scan = self._experiments.scans()[0]
        try:
            images_per_rad = 1.0 / abs(scan.get_oscillation(deg=False)[1])
        except AttributeError:
            images_per_rad = None

        for idetector, detector in enumerate(self._experiments.detectors()):
            if len(detector) == 1:
                continue
            logger.info("\nDetector %s RMSDs by panel:", idetector + 1)

            header = ["Panel\nid", "Nref"]
            for (name, units) in zip(self._target.rmsd_names, self._target.rmsd_units):
                if name == "RMSD_X" or name == "RMSD_Y" and units == "mm":
                    header.append(name + "\n(px)")
                elif (
                    name == "RMSD_Phi" and units == "rad"
                ):  # convert radians to images for reporting of scans
                    header.append("RMSD_Z" + "\n(images)")
                elif (
                    name == "RMSD_DeltaPsi" and units == "rad"
                ):  # convert radians to degrees for reporting of stills
                    header.append(name + "\n(deg)")
                else:  # skip RMSDs that cannot be expressed in image/scan space
                    pass

            rows = []
            for ipanel, panel in enumerate(detector):

                px_size = panel.get_pixel_size()
                px_per_mm = [1.0 / e for e in px_size]
                num = self._target.get_num_matches_for_panel(ipanel)
                if num <= 0:
                    continue
                raw_rmsds = self._target.rmsds_for_panel(ipanel)
                if raw_rmsds is None:
                    continue  # skip panels where rmsd cannot be calculated
                rmsds = []
                for (name, units, rmsd) in zip(
                    self._target.rmsd_names, self._target.rmsd_units, raw_rmsds
                ):
                    if name == "RMSD_X" and units == "mm":
                        rmsds.append(rmsd * px_per_mm[0])
                    elif name == "RMSD_Y" and units == "mm":
                        rmsds.append(rmsd * px_per_mm[1])
                    elif name == "RMSD_Phi" and units == "rad":
                        rmsds.append(rmsd * images_per_rad)
                    elif name == "RMSD_DeltaPsi" and units == "rad":
                        rmsds.append(rmsd * RAD2DEG)
                rows.append([str(ipanel), str(num)] + [f"{r:.5g}" for r in rmsds])

            if len(rows) > 0:
                logger.info(dials.util.tabulate(rows, header))

        return

    def run(self):
        """Run refinement"""

        ####################################
        # Do refinement and return history #
        ####################################

        logger.debug("\nExperimental models before refinement:")
        for i, beam in enumerate(self._experiments.beams()):
            logger.debug(ordinal_number(i) + " " + str(beam))
        for i, detector in enumerate(self._experiments.detectors()):
            logger.debug(ordinal_number(i) + " " + str(detector))
        for i, goniometer in enumerate(self._experiments.goniometers()):
            if goniometer is None:
                continue
            logger.debug(ordinal_number(i) + " " + str(goniometer))
        for i, scan in enumerate(self._experiments.scans()):
            if scan is None:
                continue
            logger.debug(ordinal_number(i) + " " + str(scan))
        for i, crystal in enumerate(self._experiments.crystals()):
            logger.debug(ordinal_number(i) + " " + str(crystal))

        self._refinery.run()

        # These involve calculation, so skip them when output is quiet
        if logger.getEffectiveLevel() < logging.ERROR:
            self.print_step_table()
            self.print_out_of_sample_rmsd_table()
            self.print_exp_rmsd_table()

        det_npanels = [len(d) for d in self._experiments.detectors()]
        if any(n > 1 for n in det_npanels):
            self.print_panel_rmsd_table()

        # Perform post-run tasks to write the refined states back to the models
        self._update_models()

        logger.debug("\nExperimental models after refinement:")
        for i, beam in enumerate(self._experiments.beams()):
            logger.debug(ordinal_number(i) + " " + str(beam))
        for i, detector in enumerate(self._experiments.detectors()):
            logger.debug(ordinal_number(i) + " " + str(detector))
        for i, goniometer in enumerate(self._experiments.goniometers()):
            if goniometer is None:
                continue
            logger.debug(ordinal_number(i) + " " + str(goniometer))
        for i, scan in enumerate(self._experiments.scans()):
            if scan is None:
                continue
            logger.debug(ordinal_number(i) + " " + str(scan))
        for i, crystal in enumerate(self._experiments.crystals()):
            logger.debug(ordinal_number(i) + " " + str(crystal))

        # Report on the refined parameters
        logger.debug(str(self._param_report))

        # Return the refinement history
        return self._refinery.history

    def _update_models(self):
        """Perform any extra tasks required to update the models after refinement.
        Does nothing here, but used by subclasses"""
        pass

    def selection_used_for_refinement(self):
        """Return a selection as a flex.bool in terms of the input reflection
        data of those reflections that were used in the final step of
        refinement."""

        matches = self._refman.get_matches()
        selection = flex.bool(len(self._refman.get_indexed()), False)

        try:  # new reflection table format for matches
            isel = matches["iobs"]
            selection.set_selected(isel, True)
        except TypeError:  # old ObsPredMatch format for matches
            for m in matches:
                selection[m.iobs] = True

        return selection

    def predict_for_indexed(self):
        """perform prediction for all the indexed reflections passed into
        refinement and additionally set the used_in_refinement flag. Do not
        compose the derivatives of states of the model as this is expensive and
        they are not needed outside of a refinement run"""

        reflections = self.predict_for_reflection_table(
            self._refman.get_indexed(), skip_derivatives=True
        )
        reflections.sort("iobs")
        mask = self.selection_used_for_refinement()
        reflections.set_flags(mask, reflections.flags.used_in_refinement)
        return reflections

    def predict_for_reflection_table(self, reflections, skip_derivatives=False):
        """perform prediction for all reflections in the supplied table"""

        # delegate to the target object, which has access to the predictor
        return self._target.predict_for_reflection_table(reflections, skip_derivatives)


class ScanVaryingRefiner(Refiner):
    """Includes functionality to update the models with their states at
    scan-points after scan-varying refinement"""

    def _update_models(self):
        for iexp, exp in enumerate(self._experiments):
            ar_range = exp.scan.get_array_range()
            obs_image_numbers = list(range(ar_range[0], ar_range[1] + 1))

            # write scan-varying s0 vectors back to beam models
            s0_list = self._pred_param.get_varying_s0(obs_image_numbers, iexp)
            if s0_list is not None:
                exp.beam.set_s0_at_scan_points(s0_list)

            # write scan-varying setting rotation matrices back to goniometer models
            S_list = self._pred_param.get_varying_setting_rotation(
                obs_image_numbers, iexp
            )
            if S_list is not None:
                exp.goniometer.set_setting_rotation_at_scan_points(S_list)

            # write scan-varying crystal setting matrices back to crystal models
            A_list = self._pred_param.get_varying_UB(obs_image_numbers, iexp)
            if A_list is not None:
                exp.crystal.set_A_at_scan_points(A_list)

            # Calculate scan-varying errors if requested
            if self._pred_param.set_scan_varying_errors:

                # get state covariance matrices the whole range of images. We select
                # the first element of this at each image because crystal scan-varying
                # parameterisations are not multi-state
                state_cov_list = [
                    self._pred_param.calculate_model_state_uncertainties(
                        obs_image_number=t, experiment_id=iexp
                    )
                    for t in range(ar_range[0], ar_range[1] + 1)
                ]
                if "U_cov" in state_cov_list[0]:
                    u_cov_list = [e["U_cov"] for e in state_cov_list]
                else:
                    u_cov_list = None

                if "B_cov" in state_cov_list[0]:
                    b_cov_list = [e["B_cov"] for e in state_cov_list]
                else:
                    b_cov_list = None

                # return these to the model parameterisations to be set in the models
                self._pred_param.set_model_state_uncertainties(
                    u_cov_list, b_cov_list, iexp
                )



from __future__ import annotations

import pytest

from dxtbx.model import Detector

from dials.algorithms.refinement import DialsRefineConfigError
from dials.algorithms.refinement.parameterisation.autoreduce import AutoReduce
from dials.algorithms.refinement.parameterisation.autoreduce import (
    phil_scope as ar_phil_scope,
)
from dials.algorithms.refinement.parameterisation.detector_parameters import (
    DetectorParameterisationMultiPanel,
)
from dials.algorithms.refinement.parameterisation.prediction_parameters_stills import (
    StillsPredictionParameterisation,
)
from dials.algorithms.refinement.prediction.managed_predictors import (
    StillsExperimentsPredictor,
)
from dials.algorithms.refinement.reflection_manager import ReflectionManagerFactory
from dials.algorithms.refinement.reflection_manager import (
    phil_scope as refman_phil_scope,
)

from .test_multi_panel_detector_parameterisation import make_panel_in_array
from .test_stills_prediction_parameters import _Test


@pytest.fixture(scope="session")
def tc():
    test = _Test()

    # Predict the reflections in place and put in a reflection manager
    ref_predictor = StillsExperimentsPredictor(test.stills_experiments)
    ref_predictor(test.reflections)
    test.refman = ReflectionManagerFactory.from_parameters_reflections_experiments(
        refman_phil_scope.extract(),
        test.reflections,
        test.stills_experiments,
        do_stills=True,
    )
    test.refman.finalise()

    # Build a prediction parameterisation for the stills experiment
    test.pred_param = StillsPredictionParameterisation(
        test.stills_experiments,
        detector_parameterisations=[test.det_param],
        beam_parameterisations=[test.s0_param],
        xl_orientation_parameterisations=[test.xlo_param],
        xl_unit_cell_parameterisations=[test.xluc_param],
    )
    return test


def test_check_and_fail(tc):

    # There are 823 reflections
    assert len(tc.refman.get_matches()) == 823

    # The parameters affecting the smallest number of reflections are
    # g_param_0, g_param_3 and g_param_4, all of which have gradients for
    # 792 reflections. Setting 792 reflections as the minimum should pass.
    options = ar_phil_scope.extract()
    options.min_nref_per_parameter = 792
    ar = AutoReduce(options, pred_param=tc.pred_param, reflection_manager=tc.refman)

    ar.check_and_fail()

    # Setting 793 reflections as the minimum should fail
    options.min_nref_per_parameter = 793
    ar = AutoReduce(options, pred_param=tc.pred_param, reflection_manager=tc.refman)

    with pytest.raises(DialsRefineConfigError):
        ar.check_and_fail()


def test_check_and_fix(tc):

    n_det = tc.det_param.num_free()
    n_beam = tc.s0_param.num_free()
    n_xlo = tc.xlo_param.num_free()
    n_xluc = tc.xluc_param.num_free()

    # Similar to test_check_and_fail, setting 792 reflections as the minimum
    # should leave all parameters free
    options = ar_phil_scope.extract()
    options.min_nref_per_parameter = 792
    ar = AutoReduce(options, pred_param=tc.pred_param, reflection_manager=tc.refman)
    ar.check_and_fix()

    det_params = tc.pred_param.get_detector_parameterisations()
    beam_params = tc.pred_param.get_beam_parameterisations()
    xl_ori_params = tc.pred_param.get_crystal_orientation_parameterisations()
    xl_uc_params = tc.pred_param.get_crystal_unit_cell_parameterisations()
    assert det_params[0].num_free() == n_det == 6
    assert beam_params[0].num_free() == n_beam == 3
    assert xl_ori_params[0].num_free() == n_xlo == 3
    assert xl_uc_params[0].num_free() == n_xluc == 6

    # Setting 793 reflections as the minimum should fix crystal unit cell
    # parameters g_param_0, g_param_3 and g_param_4
    options = ar_phil_scope.extract()
    options.min_nref_per_parameter = 793
    ar = AutoReduce(options, pred_param=tc.pred_param, reflection_manager=tc.refman)
    ar.check_and_fix()

    det_params = tc.pred_param.get_detector_parameterisations()
    beam_params = tc.pred_param.get_beam_parameterisations()
    xl_ori_params = tc.pred_param.get_crystal_orientation_parameterisations()
    xl_uc_params = tc.pred_param.get_crystal_unit_cell_parameterisations()
    assert det_params[0].num_free() == n_det
    assert xl_uc_params[0].num_free() == 3
    assert xl_uc_params[0].get_param_names() == ["g_param_1", "g_param_2", "g_param_5"]
    assert beam_params[0].num_free() == n_beam
    assert xl_ori_params[0].num_free() == n_xlo


def test_check_and_remove():

    test = _Test()

    # Override the single panel model and parameterisation. This test function
    # exercises the code for non-hierarchical multi-panel detectors. The
    # hierarchical detector version is tested via test_cspad_refinement.py
    multi_panel_detector = Detector()
    for x in range(3):
        for y in range(3):
            new_panel = make_panel_in_array((x, y), test.detector[0])
            multi_panel_detector.add_panel(new_panel)
    test.detector = multi_panel_detector
    test.stills_experiments[0].detector = multi_panel_detector
    test.det_param = DetectorParameterisationMultiPanel(multi_panel_detector, test.beam)

    # update the generated reflections
    test.generate_reflections()

    # Predict the reflections in place and put in a reflection manager
    ref_predictor = StillsExperimentsPredictor(test.stills_experiments)
    ref_predictor(test.reflections)
    test.refman = ReflectionManagerFactory.from_parameters_reflections_experiments(
        refman_phil_scope.extract(),
        test.reflections,
        test.stills_experiments,
        do_stills=True,
    )
    test.refman.finalise()

    # Build a prediction parameterisation for the stills experiment
    test.pred_param = StillsPredictionParameterisation(
        test.stills_experiments,
        detector_parameterisations=[test.det_param],
        beam_parameterisations=[test.s0_param],
        xl_orientation_parameterisations=[test.xlo_param],
        xl_unit_cell_parameterisations=[test.xluc_param],
    )

    # A non-hierarchical detector does not have panel groups, thus panels are
    # not treated independently wrt which reflections affect their parameters.
    # As before, setting 792 reflections as the minimum should leave all
    # parameters free, and should not remove any reflections
    options = ar_phil_scope.extract()
    options.min_nref_per_parameter = 792
    ar = AutoReduce(options, pred_param=test.pred_param, reflection_manager=test.refman)
    ar.check_and_remove()

    det_params = test.pred_param.get_detector_parameterisations()
    beam_params = test.pred_param.get_beam_parameterisations()
    xl_ori_params = test.pred_param.get_crystal_orientation_parameterisations()
    xl_uc_params = test.pred_param.get_crystal_unit_cell_parameterisations()
    assert det_params[0].num_free() == 6
    assert beam_params[0].num_free() == 3
    assert xl_ori_params[0].num_free() == 3
    assert xl_uc_params[0].num_free() == 6
    assert len(test.refman.get_obs()) == 823

    # Setting 793 reflections as the minimum fixes 3 unit cell parameters,
    # and removes all those reflections. There are then too few reflections
    # for any parameterisation and all will be fixed, leaving no free
    # parameters for refinement. This fails within PredictionParameterisation,
    # during update so the final 31 reflections are not removed.
    options = ar_phil_scope.extract()
    options.min_nref_per_parameter = 793
    ar = AutoReduce(options, pred_param=test.pred_param, reflection_manager=test.refman)
    with pytest.raises(
        DialsRefineConfigError, match="There are no free parameters for refinement"
    ):
        ar.check_and_remove()

    det_params = test.pred_param.get_detector_parameterisations()
    beam_params = test.pred_param.get_beam_parameterisations()
    xl_ori_params = test.pred_param.get_crystal_orientation_parameterisations()
    xl_uc_params = test.pred_param.get_crystal_unit_cell_parameterisations()
    assert det_params[0].num_free() == 0
    assert beam_params[0].num_free() == 0
    assert xl_ori_params[0].num_free() == 0
    assert xl_uc_params[0].num_free() == 0
    assert len(test.refman.get_obs()) == 823 - 792



from __future__ import annotations

import logging
import math

import numpy as np

import libtbx.phil
from scitbx.array_family import flex
from scitbx.math import five_number_summary

import dials.util
from dials.algorithms.clustering.observers import uc_params_from_experiments
from dials.util import log
from dials.util.options import ArgumentParser, flatten_experiments
from dials.util.version import dials_version

logger = logging.getLogger("dials.unit_cell_histogram")

help_message = """
"""

phil_scope = libtbx.phil.parse(
    """
steps_per_angstrom = 20
  .type = int
iqr_ratio = 1.5
  .type = float
""",
    process_includes=True,
)


def panel_distances_from_experiments(experiments):
    distances = [flex.double() for i in range(len(experiments[0].detector))]
    for expt in experiments:
        for i, panel in enumerate(expt.detector):
            distances[i].append(panel.get_distance())
    return distances


def outlier_selection(uc_params, iqr_ratio=1.5):
    outliers = flex.bool(uc_params[0].size(), False)
    for p in uc_params:
        min_x, q1_x, med_x, q3_x, max_x = five_number_summary(p)
        logger.info(
            "Five number summary: min %.2f, q1 %.2f, med %.2f, q3 %.2f, max %.2f",
            min_x,
            q1_x,
            med_x,
            q3_x,
            max_x,
        )
        iqr_x = q3_x - q1_x
        if iqr_x < 1e-6:
            continue
        cut_x = iqr_ratio * iqr_x
        outliers.set_selected(p > q3_x + cut_x, True)
        outliers.set_selected(p < q1_x - cut_x, True)
    logger.info("Identified %i unit cell outliers", outliers.count(True))
    return outliers


@dials.util.show_mail_handle_errors()
def run(args=None):
    usage = "dials.unit_cell_histogram [options] models.expt"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    logger.info(dials_version())

    params, options = parser.parse_args(args, show_diff_phil=False)
    experiments = flatten_experiments(params.input.experiments)

    if len(experiments) == 0:
        parser.print_help()
        exit(0)

    log.config()

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    uc_params = uc_params_from_experiments(experiments)
    panel_distances = panel_distances_from_experiments(experiments)
    outliers = outlier_selection(uc_params, iqr_ratio=params.iqr_ratio)
    plot_uc_histograms(uc_params, outliers, params.steps_per_angstrom)
    plot_uc_vs_detector_distance(
        uc_params, panel_distances, outliers, params.steps_per_angstrom
    )
    plot_number_of_crystals(experiments)


def plot_uc_histograms(
    uc_params, outliers, steps_per_angstrom=20, plot_name="uc_histograms.png"
):
    from matplotlib import pyplot as plt

    plt.style.use("ggplot")
    f, ax = plt.subplots(nrows=2, ncols=3, figsize=(12, 8))
    a, b, c = uc_params[:3]

    def uc_param_hist2d(p1, p2, ax):
        nbins = 100
        H, xedges, yedges = np.histogram2d(p1, p2, bins=nbins)
        H = np.rot90(H)
        H = np.flipud(H)
        Hmasked = np.ma.masked_where(H == 0, H)
        ax.pcolormesh(xedges, yedges, Hmasked)

    uc_param_hist2d(a, b, ax[0][0])
    uc_param_hist2d(b, c, ax[0][1])
    uc_param_hist2d(c, a, ax[0][2])

    for i in range(3):
        mmm = flex.min_max_mean_double(uc_params[i])
        steps_per_A = steps_per_angstrom
        Amin = math.floor(mmm.min * steps_per_A) / steps_per_A
        Amax = math.floor(mmm.max * steps_per_A) / steps_per_A
        n_slots = max(1, int((Amax - Amin) * steps_per_A))
        if Amin == Amax:
            eps = 0.05
            Amin -= eps
            Amax += eps
        hist = flex.histogram(uc_params[i], Amin, Amax, n_slots=n_slots)
        hist_inliers = flex.histogram(
            uc_params[i].select(~outliers), Amin, Amax, n_slots=n_slots
        )
        ax[1][i].bar(
            hist.slot_centers(),
            hist.slots(),
            align="center",
            width=hist.slot_width(),
            zorder=10,
            color="black",
            edgecolor=None,
            linewidth=0,
        )
        ax[1][i].bar(
            hist_inliers.slot_centers(),
            hist_inliers.slots(),
            align="center",
            width=hist_inliers.slot_width(),
            zorder=10,
            color="red",
            edgecolor=None,
            linewidth=0,
        )
        ax[0][i].set_xlim(ax[1][i].get_xlim())

    ax[0][0].set_ylabel(r"b ($\AA$)")
    ax[0][1].set_ylabel(r"c ($\AA$)")
    ax[0][2].set_ylabel(r"a ($\AA$)")
    ax[1][0].set_xlabel(r"a ($\AA$)")
    ax[1][1].set_xlabel(r"b ($\AA$)")
    ax[1][2].set_xlabel(r"c ($\AA$)")

    f.savefig(plot_name)
    plt.tight_layout()
    plt.close(f)


def plot_uc_vs_detector_distance(
    uc_params,
    panel_distances,
    outliers,
    steps_per_angstrom=20,
    filename="uc_vs_distance.png",
):
    from matplotlib import pyplot as plt

    plt.style.use("ggplot")
    f = plt.figure(figsize=(12, 8))
    ax1 = plt.subplot2grid((2, 3), (0, 0))
    ax2 = plt.subplot2grid((2, 3), (0, 1), sharey=ax1)
    ax3 = plt.subplot2grid((2, 3), (0, 2), sharey=ax1)
    ax4 = plt.subplot2grid((2, 3), (1, 0), colspan=3)
    a, b, c = uc_params[:3]

    def hist2d(p1, p2, ax):
        nbins = 100
        H, xedges, yedges = np.histogram2d(p1, p2, bins=nbins)
        H = np.rot90(H)
        H = np.flipud(H)
        Hmasked = np.ma.masked_where(H == 0, H)
        ax.pcolormesh(xedges, yedges, Hmasked)

    hist2d(a, panel_distances[0], ax1)
    hist2d(b, panel_distances[0], ax2)
    hist2d(c, panel_distances[0], ax3)

    mmm = flex.min_max_mean_double(panel_distances[0])
    steps_per_mm = 20
    Amin = math.floor(mmm.min * steps_per_mm) / steps_per_mm
    Amax = math.floor(mmm.max * steps_per_mm) / steps_per_mm
    n_slots = max(1, int((Amax - Amin) * steps_per_mm))
    if Amin == Amax:
        eps = 0.05
        Amin -= eps
        Amax += eps
    hist = flex.histogram(panel_distances[0], Amin, Amax, n_slots=n_slots)
    ax4.bar(
        hist.slot_centers(),
        hist.slots(),
        align="center",
        width=hist.slot_width(),
        zorder=10,
        color="red",
        edgecolor=None,
        linewidth=0,
    )

    ax1.set_ylabel("Detector distance (mm)")
    ax1.set_xlabel(r"a ($\AA$)")
    ax2.set_xlabel(r"b ($\AA$)")
    ax3.set_xlabel(r"c ($\AA$)")
    ax4.set_xlabel("Detector distance (mm)")

    f.savefig(filename)
    plt.tight_layout()
    f.clf()


def plot_number_of_crystals(experiments):
    image_to_expts = {}
    for expt in experiments:
        img = expt.imageset.get_image_identifier(0)
        image_to_expts.setdefault(img, [])
        image_to_expts[img].append(expt)

    n_crystals_per_image = flex.int(len(expts) for expts in image_to_expts.values())
    nmax = flex.max(n_crystals_per_image)
    hist = flex.histogram(n_crystals_per_image.as_double(), 0, nmax, n_slots=nmax)
    # hist.show()
    from matplotlib import pyplot as plt

    plt.style.use("ggplot")
    plt.bar(
        hist.slot_centers(),
        hist.slots(),
        align="center",
        width=hist.slot_width(),
        zorder=10,
        color="black",
        edgecolor=None,
    )
    plt.savefig("n_crystals_hist.png")
    plt.clf()


if __name__ == "__main__":
    run()



from __future__ import annotations

from cctbx import miller
from dxtbx.model import ExperimentList

from dials.algorithms.spot_prediction import ScanStaticReflectionPredictor
from dials.util import missing_reflections


def test_connected_components(dials_data):
    experiment = ExperimentList.from_file(
        dials_data("centroid_test_data", pathlib=True) / "experiments.json"
    )[0]

    image_ranges = [(1, 9), (1, 100), (1, 1000)]
    expected_ms_sizes = [[755], [242, 14, 10, 5, 2, 2, 2], []]
    for image_range, expected_sizes in zip(image_ranges, expected_ms_sizes):
        experiment.scan.set_image_range(image_range)
        predict = ScanStaticReflectionPredictor(experiment, dmin=3, margin=1)
        refl = predict.for_ub(experiment.crystal.get_A())
        miller_set = miller.set(
            experiment.crystal.get_crystal_symmetry(),
            refl["miller_index"],
            anomalous_flag=False,
        )
        miller_array = miller_set.d_spacings().resolution_filter(d_min=3)
        complete_set, unique_ms = missing_reflections.connected_components(miller_array)
        assert len(unique_ms) == len(expected_sizes)
        assert [ms.size() for ms in unique_ms] == expected_sizes
        # Verify that all the indices reported missing are actually missing from the input
        for ms in unique_ms:
            assert ms.common_set(miller_array.map_to_asu()).size() == 0
        assert complete_set.completeness() == 1


def test_connected_components_centred_cell(dials_data):
    experiment = ExperimentList.from_file(
        dials_data("insulin_processed", pathlib=True) / "scaled.expt",
        check_format=False,
    )[0]

    experiment.scan.set_image_range((1, 10))
    predict = ScanStaticReflectionPredictor(experiment, dmin=3, margin=1)
    refl = predict.for_ub(experiment.crystal.get_A())
    miller_set = miller.set(
        experiment.crystal.get_crystal_symmetry(),
        refl["miller_index"],
        anomalous_flag=False,
    )
    miller_array = miller_set.d_spacings().resolution_filter(d_min=3)
    complete_set, unique_ms = missing_reflections.connected_components(miller_array)
    assert [ms.size() for ms in unique_ms] == [581, 32, 29, 6, 3, 3, 3, 2]
    # Verify that all the indices reported missing are actually missing from the input
    for ms in unique_ms:
        assert ms.common_set(miller_array.map_to_asu()).size() == 0
    assert complete_set.completeness() == 1



from __future__ import annotations

import collections
import math

from cctbx import sgtbx, uctbx
from libtbx.math_utils import nearest_integer as nint
from scitbx import matrix

from dials.algorithms.integration import filtering
from dials.array_family import flex
from dials.util import tabulate

Slot = collections.namedtuple("Slot", "d_min d_max")
_stats_field_names = [
    "d_min_distl_method_1",
    "d_min_distl_method_2",
    "estimated_d_min",
    "n_spots_4A",
    "n_spots_no_ice",
    "n_spots_total",
    "noisiness_method_1",
    "noisiness_method_2",
    "total_intensity",
]
StatsSingleImage = collections.namedtuple("StatsSingleImage", _stats_field_names)


class StatsMultiImage(collections.namedtuple("StatsMultiImage", _stats_field_names)):
    __slots__ = ()

    def as_table(self, perm=None, n_rows=None):
        if hasattr(self, "image"):
            image = self.image
        else:
            image = flex.int(range(1, len(self.n_spots_total) + 1)).as_string()

        rows = [["image", "#spots", "#spots_no_ice", "total_intensity"]]

        estimated_d_min = None
        d_min_distl_method_1 = None
        d_min_distl_method_2 = None
        n_indexed = getattr(self, "n_indexed", None)
        fraction_indexed = getattr(self, "fraction_indexed", None)

        if flex.double(self.estimated_d_min).all_gt(0):
            estimated_d_min = self.estimated_d_min
            rows[0].append("d_min")
        if flex.double(self.d_min_distl_method_1).all_gt(0):
            d_min_distl_method_1 = self.d_min_distl_method_1
            rows[0].append("d_min (distl method 1)")
        if flex.double(self.d_min_distl_method_2).all_gt(0):
            d_min_distl_method_2 = self.d_min_distl_method_2
            rows[0].append("d_min (distl method 2)")
        if n_indexed is not None:
            rows[0].append("#indexed")
        if fraction_indexed is not None:
            rows[0].append("fraction_indexed")

        if perm is None:
            perm = list(range(len(self.n_spots_total)))
        if n_rows is not None:
            n_rows = min(n_rows, len(perm))
            perm = perm[:n_rows]
        for i_image in perm:
            d_min_str = ""
            method1_str = ""
            method2_str = ""
            if self.estimated_d_min is not None and self.estimated_d_min[i_image] > 0:
                d_min_str = f"{self.estimated_d_min[i_image]:.2f}"
            if (
                self.d_min_distl_method_1 is not None
                and self.d_min_distl_method_1[i_image] > 0
            ):
                method1_str = f"{self.d_min_distl_method_1[i_image]:.2f}"
                if self.noisiness_method_1 is not None:
                    method1_str += f" ({self.noisiness_method_1[i_image]:.2f})"
            if (
                self.d_min_distl_method_2 is not None
                and self.d_min_distl_method_2[i_image] > 0
            ):
                method2_str = f"{self.d_min_distl_method_2[i_image]:.2f}"
                if self.noisiness_method_2 is not None:
                    method2_str += f" ({self.noisiness_method_2[i_image]:.2f})"
            row = [
                image[i_image],
                str(self.n_spots_total[i_image]),
                str(self.n_spots_no_ice[i_image]),
                f"{self.total_intensity[i_image]:.0f}",
            ]
            if estimated_d_min is not None:
                row.append(d_min_str)
            if d_min_distl_method_1 is not None:
                row.append(method1_str)
            if d_min_distl_method_2 is not None:
                row.append(method2_str)
            if n_indexed is not None:
                row.append("%i" % self.n_indexed[i_image])
            if fraction_indexed is not None:
                row.append(f"{self.fraction_indexed[i_image]:.2f}")
            rows.append(row)
        return rows

    def __str__(self):
        return tabulate(self.as_table(), headers="firstrow")


class binner_equal_population:
    def __init__(self, d_star_sq, target_n_per_bin=20, max_slots=20, min_slots=5):
        n_slots = len(d_star_sq) // target_n_per_bin
        if max_slots is not None:
            n_slots = min(n_slots, max_slots)
        if min_slots is not None:
            n_slots = max(n_slots, min_slots)
        self.bins = []
        n_per_bin = len(d_star_sq) / n_slots
        d_star_sq_sorted = flex.sorted(d_star_sq)
        d_sorted = uctbx.d_star_sq_as_d(d_star_sq_sorted)
        d_max = d_sorted[0]
        for i in range(n_slots):
            d_min = d_sorted[nint((i + 1) * n_per_bin) - 1]
            self.bins.append(Slot(d_min, d_max))
            d_max = d_min


class binner_d_star_cubed:
    def __init__(self, d_spacings, target_n_per_bin=25, max_slots=40, min_slots=20):
        d_spacings = flex.double(list(set(d_spacings)))
        d_spacings_sorted = flex.sorted(d_spacings, reverse=True)
        d_star_cubed_sorted = flex.pow(1 / d_spacings_sorted, 3)

        # choose bin volume such that lowest resolution shell contains 5% of the
        # spots, or 25, whichever is greater
        low_res_count = int(
            math.ceil(
                min(
                    max(target_n_per_bin, 0.05 * len(d_spacings)),
                    0.25 * len(d_spacings),
                )
            )
        )
        bin_step = d_star_cubed_sorted[low_res_count] - d_star_cubed_sorted[0]
        assert bin_step > 0
        n_slots = int(
            math.ceil((d_star_cubed_sorted[-1] - d_star_cubed_sorted[0]) / bin_step)
        )

        if max_slots is not None:
            n_slots = min(n_slots, max_slots)
        if min_slots is not None:
            n_slots = max(n_slots, min_slots)
        bin_step = (d_star_cubed_sorted[-1] - d_star_cubed_sorted[0]) / n_slots

        self.bins = []
        ds3_max = d_star_cubed_sorted[0]
        for i in range(n_slots):
            ds3_min = d_star_cubed_sorted[0] + (i + 1) * bin_step
            self.bins.append(Slot(1 / ds3_min ** (1 / 3), 1 / ds3_max ** (1 / 3)))
            ds3_max = ds3_min


def outlier_rejection(reflections):
    # http://scripts.iucr.org/cgi-bin/paper?ba0032
    if len(reflections) == 1:
        return reflections
    intensities = reflections["intensity.sum.value"]
    variances = reflections["intensity.sum.variance"]

    i_max = flex.max_index(intensities)

    sel = flex.bool(len(reflections), True)
    sel[i_max] = False

    i_test = intensities[i_max]
    var_test = variances[i_max]

    intensities_subset = intensities.select(sel)
    var_subset = variances.select(sel)

    var_prior = var_test + 1 / flex.sum(1 / var_subset)
    p_prior = (
        1
        / math.sqrt(2 * math.pi * var_prior)
        * math.exp(-((i_test - flex.mean(intensities_subset)) ** 2) / (2 * var_prior))
    )

    if p_prior > 1e-10:
        return reflections

    return outlier_rejection(reflections.select(sel))


def wilson_outliers(reflections, ice_sel=None, p_cutoff=1e-2):
    # http://scripts.iucr.org/cgi-bin/paper?ba0032
    if ice_sel is None:
        ice_sel = flex.bool(len(reflections), False)

    E_cutoff = math.sqrt(-math.log(p_cutoff))
    intensities = reflections["intensity.sum.value"]

    Sigma_n = flex.mean(intensities.select(~ice_sel))
    normalised_amplitudes = flex.sqrt(intensities) / math.sqrt(Sigma_n)

    outliers = normalised_amplitudes >= E_cutoff

    if outliers.count(True):
        # iterative outlier rejection
        inliers = ~outliers
        outliers.set_selected(
            inliers,
            wilson_outliers(reflections.select(inliers), ice_sel.select(inliers)),
        )

    return outliers


def estimate_resolution_limit(reflections, ice_sel=None, plot_filename=None):
    if ice_sel is None:
        ice_sel = flex.bool(len(reflections), False)

    d_star_sq = flex.pow2(reflections["rlp"].norms())
    d_spacings = uctbx.d_star_sq_as_d(d_star_sq)

    intensities = reflections["intensity.sum.value"]
    variances = reflections["intensity.sum.variance"]

    sel = variances > 0
    intensities = intensities.select(sel)
    variances = variances.select(sel)
    ice_sel = ice_sel.select(sel)

    i_over_sigi = intensities / flex.sqrt(variances)
    log_i_over_sigi = flex.log(i_over_sigi)

    fit = flex.linear_regression(
        d_star_sq.select(~ice_sel), log_i_over_sigi.select(~ice_sel)
    )
    m = fit.slope()
    c = fit.y_intercept()

    log_i_sigi_lower = flex.double()
    d_star_sq_lower = flex.double()
    log_i_sigi_upper = flex.double()
    d_star_sq_upper = flex.double()

    binner = binner_equal_population(
        d_star_sq, target_n_per_bin=20, max_slots=20, min_slots=5
    )

    outliers_all = flex.bool(len(reflections), False)

    low_percentile_limit = 0.1
    upper_percentile_limit = 1 - low_percentile_limit
    for i_slot, slot in enumerate(binner.bins):
        sel_all = (d_spacings < slot.d_max) & (d_spacings >= slot.d_min)
        sel = ~(ice_sel) & sel_all

        if sel.count(True) == 0:
            continue

        outliers = wilson_outliers(
            reflections.select(sel_all), ice_sel=ice_sel.select(sel_all)
        )
        outliers_all.set_selected(sel_all, outliers)

        isel = sel_all.iselection().select(~(outliers) & ~(ice_sel).select(sel_all))
        log_i_over_sigi_sel = log_i_over_sigi.select(isel)
        d_star_sq_sel = d_star_sq.select(isel)

        perm = flex.sort_permutation(log_i_over_sigi_sel)
        i_lower = perm[int(math.floor(low_percentile_limit * len(perm)))]
        i_upper = perm[int(math.floor(upper_percentile_limit * len(perm)))]
        log_i_sigi_lower.append(log_i_over_sigi_sel[i_lower])
        log_i_sigi_upper.append(log_i_over_sigi_sel[i_upper])
        d_star_sq_upper.append(d_star_sq_sel[i_lower])
        d_star_sq_lower.append(d_star_sq_sel[i_upper])

    fit_upper = flex.linear_regression(d_star_sq_upper, log_i_sigi_upper)
    m_upper = fit_upper.slope()
    c_upper = fit_upper.y_intercept()
    fit_lower = flex.linear_regression(d_star_sq_lower, log_i_sigi_lower)
    m_lower = fit_lower.slope()
    c_lower = fit_lower.y_intercept()

    if m_upper == m_lower:
        intersection = (-1, -1)
        resolution_estimate = -1
        inside = flex.bool(len(d_star_sq), False)

    else:
        # http://en.wikipedia.org/wiki/Line%E2%80%93line_intersection#Given_the_equations_of_the_lines
        # with:
        # a_ = m_upper
        # b_ = m_lower
        # c_ = c_upper
        # d_ = c_lower
        # intersection == ((d_ - c_) / (a_ - b_), (a_ * d_ - b_ * c_) / (a_ - b_))
        intersection = (
            (c_lower - c_upper) / (m_upper - m_lower),
            (m_upper * c_lower - m_lower * c_upper) / (m_upper - m_lower),
        )

        inside = points_below_line(d_star_sq, log_i_over_sigi, m_upper, c_upper)
        inside = inside & ~outliers_all & ~ice_sel

        if inside.count(True) > 0:
            d_star_sq_estimate = flex.max(d_star_sq.select(inside))
            resolution_estimate = uctbx.d_star_sq_as_d(d_star_sq_estimate)
        else:
            resolution_estimate = -1

    if plot_filename is not None:
        from matplotlib import pyplot

        from dials.util.matplotlib_utils import resolution_formatter

        fig = pyplot.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.scatter(d_star_sq, log_i_over_sigi, marker="+")
        ax.scatter(
            d_star_sq.select(inside),
            log_i_over_sigi.select(inside),
            marker="+",
            color="green",
        )
        ax.scatter(
            d_star_sq.select(ice_sel),
            log_i_over_sigi.select(ice_sel),
            marker="+",
            color="black",
        )
        ax.scatter(
            d_star_sq.select(outliers_all),
            log_i_over_sigi.select(outliers_all),
            marker="+",
            color="grey",
        )
        ax.scatter(d_star_sq_upper, log_i_sigi_upper, marker="+", color="red")
        ax.scatter(d_star_sq_lower, log_i_sigi_lower, marker="+", color="red")

        if intersection[0] <= ax.get_xlim()[1] and intersection[1] <= ax.get_ylim()[1]:
            ax.scatter(
                [intersection[0]], [intersection[1]], marker="x", s=50, color="b"
            )
        xlim = pyplot.xlim()
        ax.plot(xlim, [(m * x + c) for x in xlim])
        ax.plot(xlim, [(m_upper * x + c_upper) for x in xlim], color="red")
        ax.plot(xlim, [(m_lower * x + c_lower) for x in xlim], color="red")
        ax.set_xlabel("d_star_sq")
        ax.set_ylabel("ln(I/sigI)")
        ax.set_xlim((max(-xlim[1], -0.05), xlim[1]))
        ax.set_ylim((0, ax.get_ylim()[1]))

        for i_slot, slot in enumerate(binner.bins):
            if i_slot == 0:
                ax.vlines(
                    uctbx.d_as_d_star_sq(slot.d_max),
                    0,
                    ax.get_ylim()[1],
                    linestyle="dotted",
                    color="grey",
                )
            ax.vlines(
                uctbx.d_as_d_star_sq(slot.d_min),
                0,
                ax.get_ylim()[1],
                linestyle="dotted",
                color="grey",
            )

        ax_ = ax.twiny()  # ax2 is responsible for "top" axis and "right" axis
        ax_.set_xticks(ax.get_xticks())
        ax_.set_xlim(ax.get_xlim())
        ax_.set_xlabel(r"Resolution ($\AA$)")
        ax_.xaxis.set_major_formatter(resolution_formatter)
        pyplot.savefig(plot_filename)
        pyplot.close()

    return resolution_estimate


def estimate_resolution_limit_distl_method1(reflections, plot_filename=None):
    # Implementation of Method 1 (section 2.4.4) of:
    # Z. Zhang, N. K. Sauter, H. van den Bedem, G. Snell and A. M. Deacon
    # J. Appl. Cryst. (2006). 39, 112-119
    # https://doi.org/10.1107/S0021889805040677

    variances = reflections["intensity.sum.variance"]

    sel = variances > 0
    reflections = reflections.select(sel)
    d_star_sq = flex.pow2(reflections["rlp"].norms())
    d_spacings = uctbx.d_star_sq_as_d(d_star_sq)
    d_star_cubed = flex.pow(reflections["rlp"].norms(), 3)

    step = 2
    while len(reflections) / step > 40:
        step += 1

    order = flex.sort_permutation(d_spacings, reverse=True)

    ds3_subset = flex.double()
    d_subset = flex.double()
    for i in range(len(reflections) // step):
        ds3_subset.append(d_star_cubed[order[i * step]])
        d_subset.append(d_spacings[order[i * step]])

    x = flex.double(range(len(ds3_subset)))

    # (i)
    # Usually, Pm is the last point, that is, m = n. But m could be smaller than
    # n if an unusually high number of spots are detected around a certain
    # intermediate resolution. In that case, our search for the image resolution
    # does not go outside the spot 'bump;. This is particularly useful when
    # ice-rings are present.

    slopes = (ds3_subset[1:] - ds3_subset[0]) / (x[1:] - x[0])
    skip_first = 3
    p_m = flex.max_index(slopes[skip_first:]) + 1 + skip_first

    # (ii)

    x1 = matrix.col((0, ds3_subset[0]))
    x2 = matrix.col((p_m, ds3_subset[p_m]))

    gaps = flex.double([0])
    v = matrix.col(((x2[1] - x1[1]), -(x2[0] - x1[0]))).normalize()

    for i in range(1, p_m):
        x0 = matrix.col((i, ds3_subset[i]))
        r = x1 - x0
        g = abs(v.dot(r))
        gaps.append(g)

    mv = flex.mean_and_variance(gaps)
    s = mv.unweighted_sample_standard_deviation()

    # (iii)

    p_k = flex.max_index(gaps)
    g_k = gaps[p_k]
    p_g = p_k
    for i in range(p_k + 1, len(gaps)):
        g_i = gaps[i]
        if g_i > (g_k - 0.5 * s):
            p_g = i

    d_g = d_subset[p_g]

    noisiness = 0
    n = len(ds3_subset)
    for i in range(n - 1):
        for j in range(i + 1, n - 1):
            if slopes[i] >= slopes[j]:
                noisiness += 1
    noisiness /= (n - 1) * (n - 2) / 2

    if plot_filename is not None:
        from matplotlib import pyplot

        fig = pyplot.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.scatter(range(len(ds3_subset)), ds3_subset)
        ax.set_ylabel("D^-3")
        xlim = pyplot.xlim()
        ylim = pyplot.ylim()
        ax.vlines(p_g, ylim[0], ylim[1], colors="red")
        pyplot.xlim(0, xlim[1])
        pyplot.ylim(0, ylim[1])
        pyplot.savefig(plot_filename)
        pyplot.close()

    return d_g, noisiness


def estimate_resolution_limit_distl_method2(reflections, plot_filename=None):
    # Implementation of Method 2 (section 2.4.4) of:
    # Z. Zhang, N. K. Sauter, H. van den Bedem, G. Snell and A. M. Deacon
    # J. Appl. Cryst. (2006). 39, 112-119
    # https://doi.org/10.1107/S0021889805040677

    variances = reflections["intensity.sum.variance"]

    sel = variances > 0
    reflections = reflections.select(sel)
    d_star_sq = flex.pow2(reflections["rlp"].norms())
    d_spacings = uctbx.d_star_sq_as_d(d_star_sq)

    binner = binner_d_star_cubed(d_spacings)

    bin_counts = flex.size_t()

    for i_slot, slot in enumerate(binner.bins):
        sel_all = (d_spacings < slot.d_max) & (d_spacings >= slot.d_min)
        sel = sel_all

        bin_counts.append(sel.count(True))

    # print list(bin_counts)
    t0 = (bin_counts[0] + bin_counts[1]) / 2

    mu = 0.15

    for i in range(len(bin_counts) - 1):
        tj = bin_counts[i]
        tj1 = bin_counts[i + 1]
        if (tj < (mu * t0)) and (tj1 < (mu * t0)):
            break

    d_min = binner.bins[i].d_min
    noisiness = 0
    m = len(bin_counts)
    for i in range(m):
        for j in range(i + 1, m):
            if bin_counts[i] <= bin_counts[j]:
                noisiness += 1
    noisiness /= 0.5 * m * (m - 1)

    if plot_filename is not None:
        from matplotlib import pyplot

        fig = pyplot.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.scatter(range(len(bin_counts)), bin_counts)
        ax.set_ylabel("number of spots in shell")
        xlim = pyplot.xlim()
        ylim = pyplot.ylim()
        ax.vlines(i, ylim[0], ylim[1], colors="red")
        pyplot.xlim(0, xlim[1])
        pyplot.ylim(0, ylim[1])
        pyplot.savefig(plot_filename)
        pyplot.close()

    return d_min, noisiness


def points_below_line(d_star_sq, log_i_over_sigi, m, c):
    p1 = matrix.col((0, c))
    p2 = matrix.col((1, m * 1 + c))

    def side(p1, p2, p):
        diff = p2 - p1
        perp = matrix.col((-diff[1], diff[0]))
        d = (p - p1).dot(perp)
        return math.copysign(1, d)

    inside = flex.bool(len(d_star_sq), False)
    for i, (x, y) in enumerate(zip(d_star_sq, log_i_over_sigi)):
        p = matrix.col((x, y))
        if side(p1, p2, p) < 0:
            inside[i] = True

    return inside


def ice_rings_selection(reflections, width=0.004):
    d_star_sq = flex.pow2(reflections["rlp"].norms())
    d_spacings = uctbx.d_star_sq_as_d(d_star_sq)

    unit_cell = uctbx.unit_cell((4.498, 4.498, 7.338, 90, 90, 120))
    space_group = sgtbx.space_group_info(number=194).group()

    if d_spacings:
        ice_filter = filtering.PowderRingFilter(
            unit_cell, space_group, flex.min(d_spacings), width
        )

        ice_sel = ice_filter(d_spacings)

        return ice_sel
    else:
        return None


def stats_for_reflection_table(
    reflections, resolution_analysis=True, filter_ice=True, ice_rings_width=0.004
):
    assert "rlp" in reflections, "Reflections must have been mapped to reciprocal space"
    reflections = reflections.select(reflections["rlp"].norms() > 0)

    d_star_sq = flex.pow2(reflections["rlp"].norms())
    d_spacings = uctbx.d_star_sq_as_d(d_star_sq)

    reflections_all = reflections
    reflections_no_ice = reflections_all
    ice_sel = None
    if reflections.size() and filter_ice:
        ice_sel = ice_rings_selection(reflections_all, width=ice_rings_width)
        if ice_sel is not None:
            reflections_no_ice = reflections_all.select(~ice_sel)
    n_spots_total = len(reflections_all)
    n_spots_no_ice = len(reflections_no_ice)
    n_spot_4A = (d_spacings > 4).count(True)
    intensities = reflections_no_ice["intensity.sum.value"]
    total_intensity = flex.sum(intensities)
    if resolution_analysis and n_spots_no_ice > 10:
        estimated_d_min = estimate_resolution_limit(reflections_all, ice_sel=ice_sel)
        (
            d_min_distl_method_1,
            noisiness_method_1,
        ) = estimate_resolution_limit_distl_method1(reflections_all)
        (
            d_min_distl_method_2,
            noisiness_method_2,
        ) = estimate_resolution_limit_distl_method2(reflections_all)
    else:
        estimated_d_min = -1.0
        d_min_distl_method_1 = -1.0
        noisiness_method_1 = -1.0
        d_min_distl_method_2 = -1.0
        noisiness_method_2 = -1.0

    return StatsSingleImage(
        n_spots_total=n_spots_total,
        n_spots_no_ice=n_spots_no_ice,
        n_spots_4A=n_spot_4A,
        total_intensity=total_intensity,
        estimated_d_min=estimated_d_min,
        d_min_distl_method_1=d_min_distl_method_1,
        noisiness_method_1=noisiness_method_1,
        d_min_distl_method_2=d_min_distl_method_2,
        noisiness_method_2=noisiness_method_2,
    )


def stats_per_image(experiment, reflections, resolution_analysis=True):
    n_spots_total = []
    n_spots_no_ice = []
    n_spots_4A = []
    total_intensity = []
    estimated_d_min = []
    d_min_distl_method_1 = []
    d_min_distl_method_2 = []
    noisiness_method_1 = []
    noisiness_method_2 = []

    image_number = reflections["xyzobs.px.value"].parts()[2]
    image_number = flex.floor(image_number)

    try:
        start, end = experiment.scan.get_array_range()
    except AttributeError:
        start, end = 0, 1
    for i in range(start, end):
        stats = stats_for_reflection_table(
            reflections.select(image_number == i),
            resolution_analysis=resolution_analysis,
        )
        n_spots_total.append(stats.n_spots_total)
        n_spots_no_ice.append(stats.n_spots_no_ice)
        n_spots_4A.append(stats.n_spots_4A)
        total_intensity.append(stats.total_intensity)
        estimated_d_min.append(stats.estimated_d_min)
        d_min_distl_method_1.append(stats.d_min_distl_method_1)
        noisiness_method_1.append(stats.noisiness_method_1)
        d_min_distl_method_2.append(stats.d_min_distl_method_2)
        noisiness_method_2.append(stats.noisiness_method_2)

    return StatsMultiImage(
        n_spots_total=n_spots_total,
        n_spots_no_ice=n_spots_no_ice,
        n_spots_4A=n_spots_4A,
        total_intensity=total_intensity,
        estimated_d_min=estimated_d_min,
        d_min_distl_method_1=d_min_distl_method_1,
        noisiness_method_1=noisiness_method_1,
        d_min_distl_method_2=d_min_distl_method_2,
        noisiness_method_2=noisiness_method_2,
    )


def plot_stats(stats, filename="per_image_analysis.png"):
    from matplotlib import pyplot

    n_spots_total = flex.int(stats.n_spots_total)
    n_spots_no_ice = stats.n_spots_no_ice
    n_spots_4A = stats.n_spots_4A
    estimated_d_min = flex.double(stats.estimated_d_min)
    d_min_distl_method_1 = flex.double(stats.d_min_distl_method_1)
    d_min_distl_method_2 = flex.double(stats.d_min_distl_method_2)
    total_intensity = stats.total_intensity

    i_image = flex.int(list(range(1, len(n_spots_total) + 1)))

    _, (ax1, ax2, ax3) = pyplot.subplots(nrows=3)
    ax1.scatter(
        list(i_image),
        list(n_spots_total),
        s=5,
        color="orange",
        marker="o",
        alpha=0.4,
        label="#spots (total)",
    )
    if n_spots_4A is not None:
        ax1.scatter(
            list(i_image),
            n_spots_4A,
            s=5,
            color="green",
            marker="o",
            alpha=0.4,
            label="#spots (to 4\u00c5)",
        )
    ax1.scatter(
        list(i_image),
        n_spots_no_ice,
        s=5,
        color="blue",
        marker="o",
        alpha=0.4,
        label="#spots (no ice)",
    )
    ax1.set_xlabel("Image #")
    ax1.set_ylabel("# spots")
    ax1.set_xlim((0.0, len(n_spots_total)))
    ax1.set_ylim(bottom=-0.2)
    ax1.legend(bbox_to_anchor=(1.05, 0.5), loc="center left", borderaxespad=0.0)

    sel = (estimated_d_min < 50.0) & (n_spots_total > 20) & (estimated_d_min > 0)
    ax2.scatter(
        list(i_image.select(sel)),
        list(estimated_d_min.select(sel)),
        s=10,
        color="red",
        marker="^",
        alpha=0.5,
        label="Estimated d_min",
    )
    ax2.scatter(
        list(i_image.select(sel)),
        list(d_min_distl_method_1.select(sel)),
        s=10,
        color="purple",
        marker="x",
        alpha=0.5,
        label="d_min (distl method 1)",
    )
    ax2.scatter(
        list(i_image.select(sel)),
        list(d_min_distl_method_2.select(sel)),
        s=10,
        color="grey",
        marker="+",
        alpha=0.5,
        label="d_min (distl method 2)",
    )
    ax2.set_ylabel("Resolution (\u00c5)")
    ax2.set_xlim((0, len(n_spots_total)))
    ylim = ax2.get_ylim()
    ylim = (math.floor(ylim[0]), math.ceil(ylim[1]))
    ax2.set_ylim(ylim)
    ax2.invert_yaxis()
    ax2.legend(bbox_to_anchor=(1.05, 0.5), loc="center left", borderaxespad=0.0)

    ax3.scatter(
        i_image,
        total_intensity,
        s=5,
        color="grey",
        marker="o",
        alpha=0.4,
        label="Total intensity",
    )
    ax3.set_ylabel("Total intensity")
    ax3.set_xlabel("Image #")
    ax3.legend(bbox_to_anchor=(1.05, 0.5), loc="center left", borderaxespad=0.0)

    pyplot.savefig(filename, dpi=600, bbox_inches="tight")



from __future__ import annotations

import os


def test_for_overlaps(dials_regression):
    from cctbx.array_family import flex

    from dials.algorithms.shoebox import MaskCode

    code_fgd = MaskCode.Foreground | MaskCode.Valid

    def is_fgd(code):
        return (code & code_fgd) == code_fgd

    code_bgd = MaskCode.Background | MaskCode.Valid

    code_overlap = code_fgd | code_bgd

    def is_overlap(code):
        return (code & code_overlap) == code_overlap

    from dials.util.options import Importer, flatten_experiments, flatten_reflections

    # test data
    refl_path = os.path.join(
        dials_regression,
        "integration_test_data",
        "stills_PSII",
        "idx-20161021225550223_integrated.pickle",
    )
    expt_path = os.path.join(
        dials_regression,
        "integration_test_data",
        "stills_PSII",
        "idx-20161021225550223_refined_experiments.json",
    )

    importer = Importer(
        [refl_path, refl_path, expt_path, expt_path],
        read_experiments=True,
        read_reflections=True,
        check_format=False,
    )

    reflections = flatten_reflections(importer.reflections)
    experiments = flatten_experiments(importer.experiments)

    from dials.algorithms.integration.overlaps_filter import OverlapsFilterMultiExpt

    overlaps_filter = OverlapsFilterMultiExpt(reflections[0], experiments)
    overlaps_filter.remove_foreground_foreground_overlaps()
    overlaps_filter.remove_foreground_background_overlaps()
    reflections = [overlaps_filter.refl]

    for expt, refl in zip(experiments, reflections):
        det = expt.detector
        size_fast, size_slow = det[0].get_image_size()
        mask_array = flex.size_t(size_fast * size_slow)
        for obs in refl.rows():
            shoebox = obs["shoebox"]
            fast_coords = range(shoebox.xsize())
            slow_coords = range(shoebox.ysize())
            for f, s in zip(fast_coords, slow_coords):
                f_abs = f + shoebox.bbox[0]  # relative to detector
                s_abs = s + shoebox.bbox[2]  # relative to detector
                posn = f_abs + s_abs * size_fast  # position in mask_array
                posn_in_shoebox = f + shoebox.xsize() * s  # position in shoebox
                assert not (
                    is_fgd(shoebox.mask[posn_in_shoebox]) and is_fgd(mask_array[posn])
                ), "Overlapping foreground found at indexed position (%d, %d), " % (
                    f_abs,
                    s_abs,
                ) + "observed centroid (%d, %d)" % (
                    obs["xyzcal.px"][0],
                    obs["xyzcal.px"][1],
                )
                try:
                    mask_array[posn] |= shoebox.mask[posn_in_shoebox]
                except IndexError:
                    continue
        for i, this_code in enumerate(mask_array):
            assert not is_overlap(
                this_code
            ), "Overlapping foreground and background found at (%d, %d)" % (
                i % shoebox.xsize(),
                i // shoebox.xsize(),
            )



# FIXME Mosaicity seems to be overestimated
# FIXME Don't know how XDS REFLECTING_RANGE is calculated
# FIXME Don't know what XDS REFLECTION_RANGE is used for
# FIXME Don't know what XDS BEAM_DIVERGENCE is used for
# FIXME Should maybe be scan varying
# FIXME Don't know how XDS calculated the n_sigma


from __future__ import annotations

import collections
import copy
import logging
import math

import numpy as np

import scitbx.math
from dxtbx import flumpy

from dials.array_family import flex

logger = logging.getLogger(__name__)


class ComputeEsdBeamDivergence:
    """Calculate the E.s.d of the beam divergence."""

    def __init__(self, detector, reflections, centroid_definition="s1"):
        """Calculate the E.s.d of the beam divergence.

        Params:
            detector The detector class
            reflections The reflections
            centroid_definition ENUM com or s1
        """
        # Calculate the beam direction variances
        variance = self._beam_direction_variance_list(
            detector, reflections, centroid_definition
        )

        # Calculate and return the e.s.d of the beam divergence
        self._sigma = math.sqrt(np.sum(variance) / variance.size)

    def sigma(self):
        """Return the E.S.D of the beam divergence."""
        return self._sigma

    def _beam_direction_variance_list(
        self, detector, reflections, centroid_definition="s1"
    ):
        """Calculate the variance in beam direction for each spot.

        Params:
            detector The detector model
            reflections The list of reflections
            centroid_definition ENUM com or s1

        Returns:
            The list of variances
        """
        # Get the reflection columns
        shoebox = reflections["shoebox"]
        xyz = reflections["xyzobs.px.value"]

        # Loop through all the reflections
        variance = np.array([], dtype=np.float64)

        if centroid_definition == "com":
            # Calculate the beam vector at the centroid
            s1_centroid = []
            for r in range(len(reflections)):
                panel = shoebox[r].panel
                s1_centroid.append(detector[panel].get_pixel_lab_coord(xyz[r][0:2]))
        else:
            s1_centroid = reflections["s1"]

        for r in range(len(reflections)):
            # Get the coordinates and values of valid shoebox pixels
            # FIXME maybe I note in Kabsch (2010) s3.1 step (v) is
            # background subtraction, appears to be missing here.
            mask = shoebox[r].mask != 0
            values = flumpy.to_numpy(shoebox[r].values(mask))
            s1 = shoebox[r].beam_vectors(detector, mask)

            angles = flumpy.to_numpy(s1.angle(s1_centroid[r], deg=False))

            if np.sum(values) > 1:
                var = np.sum(values * np.square(angles)) / (np.sum(values) - 1)
                variance = np.append(variance, var)

        # Return a list of variances
        return variance


class FractionOfObservedIntensity:
    """Calculate the fraction of observed intensity for different sigma_m."""

    def __init__(self, crystal, beam, detector, goniometer, scan, reflections):
        """Initialise the algorithm. Calculate the list of tau and zetas.

        Params:
            reflections The list of reflections
            experiment The experiment object
        """
        # Get the oscillation width
        dphi2 = scan.get_oscillation(deg=False)[1] / 2.0

        # Calculate a list of angles and zeta's
        tau, zeta = self._calculate_tau_and_zeta(
            crystal, beam, detector, goniometer, scan, reflections
        )

        # Calculate zeta * (tau +- dphi / 2) / math.sqrt(2)
        self.e1 = (tau + dphi2) * flex.abs(zeta) / math.sqrt(2.0)
        self.e2 = (tau - dphi2) * flex.abs(zeta) / math.sqrt(2.0)

    def _calculate_tau_and_zeta(
        self, crystal, beam, detector, goniometer, scan, reflections
    ):
        """Calculate the list of tau and zeta needed for the calculation.

        Params:
            reflections The list of reflections
            experiment The experiment object.

        Returns:
            (list of tau, list of zeta)
        """
        from dials.algorithms.shoebox import MaskCode

        mask_code = MaskCode.Valid | MaskCode.Foreground

        # Calculate the list of frames and z coords
        sbox = reflections["shoebox"]
        bbox = reflections["bbox"]
        phi = reflections["xyzcal.mm"].parts()[2]

        # Calculate the zeta list
        zeta = reflections["zeta"]

        # Calculate the list of tau values
        tau = []
        zeta2 = []
        for s, b, p, z in zip(sbox, bbox, phi, zeta):
            for z0, f in enumerate(range(b[4], b[5])):
                phi0 = scan.get_angle_from_array_index(int(f), deg=False)
                phi1 = scan.get_angle_from_array_index(int(f) + 1, deg=False)
                m = s.mask[z0 : z0 + 1, :, :]
                if m.count(mask_code) > 0:
                    tau.append((phi1 + phi0) / 2.0 - p)
                    zeta2.append(z)

        # Return the list of tau and zeta
        return flex.double(tau), flex.double(zeta2)

    def __call__(self, sigma_m):
        """Calculate the fraction of observed intensity for each observation.

        Params:
            sigma_m The mosaicity

        Returns:
            A list of log intensity fractions
        """
        # Tiny value
        TINY = 1e-10
        assert sigma_m > TINY

        # Calculate the two components to the fraction
        a = scitbx.math.erf(self.e1 / sigma_m)
        b = scitbx.math.erf(self.e2 / sigma_m)

        # Calculate the fraction of observed reflection intensity
        R = (a - b) / 2.0

        # Set any points <= 0 to 1e-10 (otherwise will get a floating
        # point error in log calculation below).
        assert R.all_ge(0)
        mask = R < TINY
        assert mask.count(True) < len(mask)
        R.set_selected(mask, TINY)

        # Return the logarithm of r
        return flex.log(R)


class ComputeEsdReflectingRange:
    """calculate the e.s.d of the reflecting range (mosaicity)."""

    class Estimator:
        """Estimate E.s.d reflecting range by maximum likelihood estimation."""

        def __init__(self, crystal, beam, detector, goniometer, scan, reflections):
            """Initialise the optimization."""
            from scitbx import simplex

            # FIXME in here this code is very unstable or actually broken if
            # we pass in a few lone images i.e. screening shots - propose need
            # for alternative algorithm, in meantime can avoid code death if
            # something like this
            #
            # if scan.get_num_images() == 1:
            #   self.sigma = 0.5 * scan.get_oscillation_range()[1] * math.pi / 180.0
            #   return
            #
            # is used... @JMP please could we discuss? assert best method to get
            # sigma_m in that case is actually to look at present / absent
            # reflections as per how Mosflm does it...
            # Initialise the function used in likelihood estimation.
            self._R = FractionOfObservedIntensity(
                crystal, beam, detector, goniometer, scan, reflections
            )

            # Set the starting values to try 1, 3 degrees seems sensible for
            # crystal mosaic spread
            start = 1 * math.pi / 180
            stop = 3 * math.pi / 180
            starting_simplex = [flex.double([start]), flex.double([stop])]

            # Initialise the optimizer
            optimizer = simplex.simplex_opt(
                1, matrix=starting_simplex, evaluator=self, tolerance=1e-7
            )

            # Get the solution
            self.sigma = math.exp(optimizer.get_solution()[0])

        def target(self, log_sigma):
            """The target for minimization."""
            return -flex.sum(self._R(math.exp(log_sigma[0])))

    class CrudeEstimator:
        """If the main estimator failed make a crude estimate"""

        def __init__(self, crystal, beam, detector, goniometer, scan, reflections):
            # Calculate a list of angles and zeta's
            tau, zeta = self._calculate_tau_and_zeta(
                crystal, beam, detector, goniometer, scan, reflections
            )

            # Calculate zeta * (tau +- dphi / 2) / math.sqrt(2)
            X = tau * zeta
            mv = flex.mean_and_variance(X)
            self.sigma = math.sqrt(mv.unweighted_sample_variance())

        def _calculate_tau_and_zeta(
            self, crystal, beam, detector, goniometer, scan, reflections
        ):
            """Calculate the list of tau and zeta needed for the calculation.

            Params:
                reflections The list of reflections
                experiment The experiment object.

            Returns:
                (list of tau, list of zeta)
            """
            from dials.algorithms.shoebox import MaskCode

            mask_code = MaskCode.Valid | MaskCode.Foreground

            # Calculate the list of frames and z coords
            sbox = reflections["shoebox"]
            bbox = reflections["bbox"]
            phi = reflections["xyzcal.mm"].parts()[2]

            # Calculate the zeta list
            zeta = reflections["zeta"]

            # Calculate the list of tau values
            tau = []
            zeta2 = []
            for s, b, p, z in zip(sbox, bbox, phi, zeta):
                for z0, f in enumerate(range(b[4], b[5])):
                    phi0 = scan.get_angle_from_array_index(int(f), deg=False)
                    phi1 = scan.get_angle_from_array_index(int(f) + 1, deg=False)
                    m = s.mask[z0 : z0 + 1, :, :]
                    if m.count(mask_code) > 0:
                        tau.append((phi1 + phi0) / 2.0 - p)
                        zeta2.append(z)

            # Return the list of tau and zeta
            return flex.double(tau), flex.double(zeta2)

    class ExtendedEstimator:
        """Try to estimate using knowledge of intensities"""

        def __init__(
            self,
            crystal,
            beam,
            detector,
            goniometer,
            scan,
            reflections,
            n_macro_cycles=10,
        ):
            from scitbx import simplex

            # Get the oscillation width
            dphi2 = scan.get_oscillation(deg=False)[1] / 2.0

            # Calculate a list of angles and zeta's
            tau, zeta, n, indices = self._calculate_tau_and_zeta(
                crystal, beam, detector, goniometer, scan, reflections
            )

            # Calculate zeta * (tau +- dphi / 2) / math.sqrt(2)
            self.e1 = (tau + dphi2) * flex.abs(zeta) / math.sqrt(2.0)
            self.e2 = (tau - dphi2) * flex.abs(zeta) / math.sqrt(2.0)
            self.indices = indices
            if len(self.e1) == 0:
                raise RuntimeError(
                    "Something went wrong. Zero pixels selected for estimation of profile parameters."
                )

            # Compute intensity
            self.K = flex.double()
            self.nj = []
            for i0, i1 in zip(self.indices[:-1], self.indices[1:]):
                nj = n[i0:i1]
                self.K.append(flex.sum(nj))
                self.nj.append(nj)

            # Set the starting values to try 1, 3 degrees seems sensible for
            # crystal mosaic spread
            start = math.log(0.1 * math.pi / 180)
            stop = math.log(1 * math.pi / 180)
            starting_simplex = [flex.double([start]), flex.double([stop])]

            # Initialise the optimizer
            optimizer = simplex.simplex_opt(
                1, matrix=starting_simplex, evaluator=self, tolerance=1e-3
            )

            # Get the solution
            sigma = math.exp(optimizer.get_solution()[0])

            # Save the result
            self.sigma = sigma

        def target(self, log_sigma):
            """The target for minimization."""
            sigma_m = math.exp(log_sigma[0])

            # Tiny value
            TINY = 1e-10
            assert sigma_m > TINY

            # Calculate the two components to the fraction
            a = scitbx.math.erf(self.e1 / sigma_m)
            b = scitbx.math.erf(self.e2 / sigma_m)

            # Calculate the fraction of observed reflection intensity
            zi = (a - b) / 2.0

            # Set any points <= 0 to 1e-10 (otherwise will get a floating
            # point error in log calculation below).
            assert zi.all_ge(0)
            mask = zi < TINY
            assert mask.count(True) < len(mask)
            zi.set_selected(mask, TINY)

            # Compute the likelihood
            #
            # The likelihood here is a result of the sum of two log likelihood
            # functions:
            #
            # The first is the same as the one in Kabsch2010 as applied to the
            # reflection as a whole. This results in the term log(Z)
            #
            # The second is the likelihood for each reflection modelling as a Poisson
            # distributions with shape given by sigma M. This gives sum(ci log(zi)) -
            # sum(ci)*log(sum(zi))
            #
            # If the reflection is recorded on 1 frame, the second component is zero
            # and so the likelihood is dominated by the first term which can be seen
            # as a prior for sigma, which accounts for which reflections were actually
            # recorded.
            #
            L = 0
            for (kj, nj, i0, i1) in zip(
                self.K, self.nj, self.indices[:-1], self.indices[1:]
            ):
                zj = zi[i0:i1]
                logZ = math.log(flex.sum(zj))
                L += flex.sum(nj * flex.log(zj)) - kj * logZ + logZ
            logger.debug("Sigma M: %f, log(L): %f", sigma_m * 180 / math.pi, L)

            # Return the logarithm of r
            return -L

        def _calculate_tau_and_zeta(
            self, crystal, beam, detector, goniometer, scan, reflections
        ):
            """Calculate the list of tau and zeta needed for the calculation.

            Params:
                reflections The list of reflections
                experiment The experiment object.

            Returns:
                (list of tau, list of zeta)
            """
            from dials.algorithms.shoebox import MaskCode

            mask_code = MaskCode.Valid | MaskCode.Foreground

            # Calculate the list of frames and z coords
            sbox = reflections["shoebox"]
            phi = reflections["xyzcal.mm"].parts()[2]

            # Calculate the zeta list
            zeta = reflections["zeta"]

            # Calculate the list of tau values
            tau = []
            zeta2 = []
            num = []
            indices = [0]
            for s, p, z in zip(sbox, phi, zeta):
                b = s.bbox
                for z0, f in enumerate(range(b[4], b[5])):
                    phi0 = scan.get_angle_from_array_index(int(f), deg=False)
                    phi1 = scan.get_angle_from_array_index(int(f) + 1, deg=False)
                    d = s.data[z0 : z0 + 1, :, :]
                    m = s.mask[z0 : z0 + 1, :, :]
                    d = flex.sum(d.as_1d().select(m.as_1d() == mask_code))
                    if d > 0:
                        tau.append((phi1 + phi0) / 2.0 - p)
                        zeta2.append(z)
                        num.append(d)
                if len(zeta2) > indices[-1]:
                    indices.append(len(zeta2))

            # Return the list of tau and zeta
            return (
                flex.double(tau),
                flex.double(zeta2),
                flex.double(num),
                flex.size_t(indices),
            )

    def __init__(
        self, crystal, beam, detector, goniometer, scan, reflections, algorithm="basic"
    ):
        """initialise the algorithm with the scan.

        params:
            scan the scan object
        """

        if algorithm == "basic":

            # Calculate sigma_m
            try:
                estimator = ComputeEsdReflectingRange.Estimator(
                    crystal, beam, detector, goniometer, scan, reflections
                )
            except Exception:
                logger.info("Using Crude Mosaicity estimator")
                estimator = ComputeEsdReflectingRange.CrudeEstimator(
                    crystal, beam, detector, goniometer, scan, reflections
                )

        elif algorithm == "extended":
            estimator = ComputeEsdReflectingRange.ExtendedEstimator(
                crystal, beam, detector, goniometer, scan, reflections
            )

        # Save the solution
        self._sigma = estimator.sigma

    def sigma(self):
        """Return the E.S.D reflecting rang."""
        return self._sigma


class ProfileModelCalculator:
    """Class to help calculate the profile model."""

    def __init__(
        self,
        reflections,
        crystal,
        beam,
        detector,
        goniometer,
        scan,
        min_zeta=0.05,
        algorithm="basic",
        centroid_definition="s1",
    ):
        """Calculate the profile model."""
        from dxtbx.model.experiment_list import Experiment

        # Check input has what we want
        assert reflections is not None
        assert "miller_index" in reflections
        assert "s1" in reflections
        assert "shoebox" in reflections
        assert "xyzobs.px.value" in reflections
        assert "xyzcal.mm" in reflections

        assert centroid_definition in ("s1", "com")

        n_all = reflections.size()

        # stills images behave differently in here
        if goniometer is None or scan is None or scan.is_still():
            logger.info("Using %d reflections for sigma calculation", n_all)
            logger.info("Calculating E.S.D Beam Divergence.")
            beam_divergence = ComputeEsdBeamDivergence(
                detector, reflections, centroid_definition
            )
            self._sigma_b = beam_divergence.sigma()
            # FIXME calculate properly
            self._sigma_m = 0.0
        else:
            # filter reflections before determining the reflection profile
            # parameters - first by used_in_refinement then by zeta

            reflections = _select_reflections_for_sigma_calc(reflections)

            zeta = reflections.compute_zeta(
                Experiment(
                    crystal=crystal,
                    beam=beam,
                    detector=detector,
                    goniometer=goniometer,
                    scan=scan,
                )
            )
            reflections = reflections.select(flex.abs(zeta) >= min_zeta)
            n_use = reflections.size()

            logger.info("Using %d / %d reflections for sigma calculation", n_use, n_all)
            logger.info("Calculating E.S.D Beam Divergence.")
            beam_divergence = ComputeEsdBeamDivergence(
                detector, reflections, centroid_definition
            )

            self._sigma_b = beam_divergence.sigma()

            logger.info("Calculating E.S.D Reflecting Range (mosaicity).")
            reflecting_range = ComputeEsdReflectingRange(
                crystal,
                beam,
                detector,
                goniometer,
                scan,
                reflections,
                algorithm=algorithm,
            )

            self._sigma_m = reflecting_range.sigma()

        logger.info(" sigma b: %f degrees", self._sigma_b * 180 / math.pi)
        logger.info(" sigma m: %f degrees", self._sigma_m * 180 / math.pi)

    def sigma_b(self):
        """Return the E.S.D beam divergence."""
        return self._sigma_b

    def sigma_m(self):
        """Return the E.S.D reflecting range."""
        return self._sigma_m


class ScanVaryingProfileModelCalculator:
    """Class to help calculate the profile model."""

    def __init__(
        self,
        reflections,
        crystal,
        beam,
        detector,
        goniometer,
        scan,
        min_zeta=0.05,
        algorithm="basic",
        centroid_definition="s1",
    ):
        """Calculate the profile model."""
        from dxtbx.model.experiment_list import Experiment

        # Check input has what we want
        assert reflections is not None
        assert "miller_index" in reflections
        assert "s1" in reflections
        assert "shoebox" in reflections
        assert "xyzobs.px.value" in reflections
        assert "xyzcal.mm" in reflections

        assert centroid_definition in ("s1", "com")

        # Select by zeta
        zeta = reflections.compute_zeta(
            Experiment(
                crystal=crystal,
                beam=beam,
                detector=detector,
                goniometer=goniometer,
                scan=scan,
            )
        )
        mask = flex.abs(zeta) >= min_zeta
        reflections = reflections.select(mask)

        # Split the reflections into partials
        reflections = copy.deepcopy(reflections)
        reflections.split_partials_with_shoebox()

        # Get a list of reflections for each frame
        bbox = reflections["bbox"]
        index_list = collections.defaultdict(list)
        for i, (x0, x1, y0, y1, z0, z1) in enumerate(bbox):
            assert z1 == z0 + 1
            index_list[z0].append(i)
        reflection_list = {
            key: reflections.select(flex.size_t(value))
            for key, value in index_list.items()
        }

        # The range of frames
        z0, z1 = scan.get_array_range()
        min_z = min(index_list)
        max_z = max(index_list)
        assert z0 == min_z
        assert z1 == max_z + 1

        # Compute for all frames
        self._num = []
        sigma_b = flex.double()
        sigma_m = flex.double()
        for i in range(z0, z1):

            # Get reflections at the index
            reflections = reflection_list[i]

            self._num.append(len(reflections))

            # Calculate the E.S.D of the beam divergence
            beam_divergence = ComputeEsdBeamDivergence(detector, reflections)

            # Set the sigma b
            sigma_b.append(beam_divergence.sigma())

            # Calculate the E.S.D of the reflecting range
            reflecting_range = ComputeEsdReflectingRange(
                crystal, beam, detector, goniometer, scan, reflections
            )

            logger.info(
                "Computing profile model for frame %d: sigma_b = %.4f degrees, sigma_m = %.4f degrees",
                i,
                beam_divergence.sigma() * 180 / math.pi,
                reflecting_range.sigma() * 180 / math.pi,
            )

            # Set the sigmas
            sigma_m.append(reflecting_range.sigma())

        def convolve(data, kernel):
            assert len(kernel) & 1
            result = []
            mid = len(kernel) // 2
            for i in range(len(data)):
                r = 0
                for j in range(len(kernel)):
                    k = i - mid + j
                    if k < 0:
                        r += kernel[j] * data[0]
                    elif k >= len(data):
                        r += kernel[j] * data[-1]
                    else:
                        r += kernel[j] * data[k]
                result.append(r)
            return result

        def gaussian_kernel(n):
            assert n & 1
            mid = n // 2
            sigma = mid / 3.0
            kernel = [math.exp(-((i - mid) ** 2) / (2 * sigma**2)) for i in range(n)]
            kernel = [k / sum(kernel) for k in kernel]
            return kernel

        # Smooth the parameters
        kernel = gaussian_kernel(51)
        sigma_b_sq_new = convolve(flex.pow2(sigma_b), kernel)
        sigma_m_sq_new = convolve(flex.pow2(sigma_m), kernel)

        # Print the output - mean as is scan varying
        mean_sigma_b = math.sqrt(sum(flex.pow2(sigma_b)) / len(sigma_b))
        mean_sigma_m = math.sqrt(sum(flex.pow2(sigma_m)) / len(sigma_m))

        # Save the smoothed parameters
        self._sigma_b = flex.sqrt(flex.double(sigma_b_sq_new))
        self._sigma_m = flex.sqrt(flex.double(sigma_m_sq_new))
        assert len(self._sigma_b) == len(self._sigma_m)

        # Print out smoothed profile parameters
        for i in range(len(sigma_b)):
            logger.info(
                "Smoothed profile model for frame %d: sigma_b = %.4f degrees, sigma_m = %.4f degrees",
                i,
                self._sigma_b[i] * 180 / math.pi,
                self._sigma_m[i] * 180 / math.pi,
            )

        # Print the mean parameters
        logger.info(" sigma b: %f degrees", mean_sigma_b * 180 / math.pi)
        logger.info(" sigma m: %f degrees", mean_sigma_m * 180 / math.pi)

    def num(self):
        """The number of reflections used."""
        return self._num

    def sigma_b(self):
        """Return the E.S.D beam divergence."""
        return self._sigma_b

    def sigma_m(self):
        """Return the E.S.D reflecting range."""
        return self._sigma_m


def _select_reflections_for_sigma_calc(reflections, min_number_of_refl=10000):
    """Determine a subset of reflections to use for sigma_m calculation."""
    n_ref = reflections.size()
    if n_ref >= min_number_of_refl:
        # ideally use well-sampled selection from refinement
        used_in_ref = reflections.get_flags(reflections.flags.used_in_refinement)
        n_used_in_ref = used_in_ref.count(True)
        if n_used_in_ref >= min_number_of_refl:
            selected_reflections = reflections.select(used_in_ref)
            logger.debug(
                "Using %s reflections with used_in_refinement flag for sigma calculation",
                n_used_in_ref,
            )

        # handle case where used_in_refl not set/data not refined
        else:
            # more than min_no of refls, but refinement didn't use more than
            # min_no: still use the ones from refinement and 'top up'
            if n_used_in_ref:
                selected_reflections = reflections.select(used_in_ref)
                other_refls = reflections.select(~used_in_ref)
            else:
                selected_reflections = flex.reflection_table()
                other_refls = reflections
            # top up with every nth reflection needed to make up the number needed
            n_sel = selected_reflections.size()
            n_other = other_refls.size()
            step_size = int(math.floor(n_other / (min_number_of_refl - n_sel)))
            sel = flex.size_t(i for i in range(0, n_other, step_size))
            selected_reflections.extend(other_refls.select(sel))
            logger.debug(
                "Using %s reflections for sigma calculation",
                selected_reflections.size(),
            )
        return selected_reflections

    logger.debug("Using all suitable reflections for sigma calculation")
    return reflections



from __future__ import annotations

from unittest import mock

import numpy as np

from dials.algorithms.symmetry.cosym import observers


def test_SymmetryAnalysisObserver():
    # setup script
    script = mock.Mock()

    # test when no symmetry analysis has been performed
    script._symmetry_analysis = None
    observer = observers.SymmetryAnalysisObserver()
    observer.update(script)
    d = observer.make_tables()
    assert d == {"symmetry_analysis": {}}

    script._symmetry_analysis = mock.Mock()
    script._symmetry_analysis.sym_ops_table = mock.Mock()
    script._symmetry_analysis.subgroups_table = mock.Mock()
    script._symmetry_analysis.as_dict = mock.Mock(
        return_value={
            "subgroup_scores": [
                {
                    "patterson_group": "-P 1",
                    "unit_cell": (10, 10, 10, 90, 90, 90),
                    "cb_op": "x,y,z",
                    "likelihood": 0.9,
                    "confidence": 0.9,
                    "stars": "*",
                    "z_cc_net": 1,
                    "z_cc_for": 2,
                    "z_cc_against": 3,
                    "max_angular_difference": 0.2,
                }
            ],
            "sym_op_scores": [
                {
                    "cc": 0.99,
                    "operator": "x,y,z",
                    "likelihood": 0.99,
                    "stars": "**",
                    "z_cc": 10,
                }
            ],
        }
    )

    # test the observer
    observer = observers.SymmetryAnalysisObserver()
    observer.update(script)
    d = observer.make_tables()
    assert "symmetry_analysis" in d
    assert set(d["symmetry_analysis"]) == {
        "summary_table",
        "subgroups_table",
        "sym_ops_table",
    }


def test_CosymClusterAnalysisObserver():
    rij_matrix = np.random.rand(16).reshape(4, 4)
    coords = np.random.rand(8).reshape(4, 2)

    # setup script
    script = mock.Mock()
    script.target = mock.Mock()
    script.target.rij_matrix = rij_matrix
    script.coords = coords
    script.cluster_labels = np.zeros(4)

    # test the observer
    observer = observers.CosymClusterAnalysisObserver()
    observer.update(script)
    d = observer.make_plots()
    assert "cosym_graphs" in d


def test_CosymHTMLGenerator():
    pass
    # script = mock.Mock()
    # script.params.output.html = "test.html"

    ## Test that CosymHTMLGenerator works if all data is empty.
    # observer = observers.CosymHTMLGenerator()
    # observer.make_html(script)
    # assert os.path.exists("test.html")



"""Methods for symmetry determination from partial datasets.

This module implements the methods of `Gildea, R. J. & Winter, G. (2018).
Acta Cryst. D74, 405-410 <https://doi.org/10.1107/S2059798318002978>`_ for
determination of Patterson group symmetry from sparse multi-crystal data sets in
the presence of an indexing ambiguity.
"""

from __future__ import annotations

import json
import logging
import math
from typing import List, Optional

import numpy as np
from sklearn.neighbors import NearestNeighbors

import iotbx.phil
from cctbx import miller, sgtbx
from cctbx.sgtbx.lattice_symmetry import metric_subgroups
from libtbx import Auto
from scitbx import matrix
from scitbx.array_family import flex

import dials.util
from dials.algorithms.indexing.symmetry import find_matching_symmetry
from dials.algorithms.symmetry import symmetry_base
from dials.algorithms.symmetry.cosym import engine as cosym_engine
from dials.algorithms.symmetry.cosym import target
from dials.algorithms.symmetry.laue_group import ScoreCorrelationCoefficient
from dials.util.observer import Subject
from dials.util.reference import intensities_from_reference_file

logger = logging.getLogger(__name__)

phil_scope = iotbx.phil.parse(
    """\

normalisation = kernel quasi *ml_iso ml_aniso
  .type = choice

d_min = Auto
  .type = float(value_min=0)

min_i_mean_over_sigma_mean = 4
  .type = float(value_min=0)
  .short_caption = "Minimum <I>/<σ>"

min_cc_half = 0.6
  .type = float(value_min=0, value_max=1)
  .short_caption = "Minimum CC½"

lattice_group = None
  .type = space_group
  .short_caption = "Lattice group"

space_group = None
  .type = space_group
  .short_caption = "Space group"

lattice_symmetry_max_delta = 5.0
  .type = float(value_min=0)
  .short_caption = "Lattice symmetry max δ"

best_monoclinic_beta = True
  .type = bool
  .help = "If True, then for monoclinic centered cells, I2 will be preferred over C2 if"
          "it gives a less oblique cell (i.e. smaller beta angle)."
  .short_caption = "Best monoclinic β"

dimensions = Auto
  .type = int(value_min=2)
  .short_caption = "Dimensions"

use_curvatures = True
  .type = bool
  .short_caption = "Use curvatures"

weights = count standard_error
  .type = choice
  .short_caption = "Weights"

min_pairs = 3
  .type = int(value_min=1)
  .help = 'Minimum number of pairs for inclusion of correlation coefficient in calculation of Rij matrix.'
  .short_caption = "Minimum number of pairs"

minimization
  .short_caption = "Minimization"
{
  engine = *scitbx scipy
    .type = choice
    .short_caption = "Engine"
  max_iterations = 100
    .type = int(value_min=0)
    .short_caption = "Maximum number of iterations"
  max_calls = None
    .type = int(value_min=0)
    .short_caption = "Maximum number of calls"
}

nproc = None
  .type = int(value_min=1)
  .help = "Deprecated"
  .deprecated = True
"""
)


class CosymAnalysis(symmetry_base, Subject):
    """Perform cosym analysis.

    Perform cosym analysis on the input intensities using the methods of
    `Gildea, R. J. & Winter, G. (2018). Acta Cryst. D74, 405-410
    <https://doi.org/10.1107/S2059798318002978>`_ for
    determination of Patterson group symmetry from sparse multi-crystal data sets in
    the presence of an indexing ambiguity.
    """

    def __init__(self, intensities, params, seed_dataset: Optional[int] = None):
        """Initialise a CosymAnalysis object.

        Args:
          intensities (cctbx.miller.array): The intensities on which to perform
            cosym analysis.
          params (libtbx.phil.scope_extract): Parameters for the analysis.
          seed_dataset (int): The index into the intensities list to use when
            choosing a seed dataset for the reindexing analysis (the x,y,z reindex
            mode will be used for this dataset).
            If None, a high density cluster point is chosen.
        """
        self.seed_dataset = seed_dataset
        if self.seed_dataset:
            self.seed_dataset = int(self.seed_dataset)
            assert (
                0 <= seed_dataset < len(intensities)
            ), "cosym_analysis: seed_dataset parameter must be an integer that can be used to index the intensities list"

        super().__init__(
            intensities,
            normalisation=params.normalisation,
            lattice_symmetry_max_delta=params.lattice_symmetry_max_delta,
            d_min=params.d_min,
            min_i_mean_over_sigma_mean=params.min_i_mean_over_sigma_mean,
            min_cc_half=params.min_cc_half,
            relative_length_tolerance=None,
            absolute_angle_tolerance=None,
            best_monoclinic_beta=params.best_monoclinic_beta,
        )
        Subject.__init__(
            self, events=["optimised", "analysed_symmetry", "analysed_clusters"]
        )

        self.params = params
        if self.params.space_group is not None:

            def _map_space_group_to_input_cell(intensities, space_group):
                from cctbx.sgtbx.bravais_types import bravais_lattice

                best_subgroup = find_matching_symmetry(
                    intensities.unit_cell(),
                    space_group,
                    best_monoclinic_beta=str(bravais_lattice(group=space_group))
                    == "mI",
                )
                cb_op_inp_best = best_subgroup["cb_op_inp_best"]
                best_subsym = best_subgroup["best_subsym"]
                cb_op_best_primitive = (
                    best_subsym.change_of_basis_op_to_primitive_setting()
                )

                sg_cb_op_inp_primitive = (
                    space_group.info().change_of_basis_op_to_primitive_setting()
                )
                sg_primitive = space_group.change_basis(sg_cb_op_inp_primitive)
                sg_best = sg_primitive.change_basis(cb_op_best_primitive.inverse())
                # best_subgroup above is the bravais type, so create thin copy here with the
                # user-input space group instead
                best_subsym = best_subsym.customized_copy(
                    space_group_info=sg_best.info()
                )
                best_subgroup = {
                    "subsym": best_subsym.change_basis(cb_op_inp_best.inverse()),
                    "best_subsym": best_subsym,
                    "cb_op_inp_best": cb_op_inp_best,
                }

                intensities = intensities.customized_copy(
                    space_group_info=sg_best.change_basis(
                        cb_op_inp_best.inverse()
                    ).info()
                )
                return intensities, best_subgroup

            self.intensities, self.best_subgroup = _map_space_group_to_input_cell(
                self.intensities, self.params.space_group.group()
            )
            self.best_subgroup["cb_op_inp_best"] = (
                self.best_subgroup["cb_op_inp_best"] * self.cb_op_inp_min
            )
            self.input_space_group = self.intensities.space_group()

        else:
            self.input_space_group = None

        if self.params.lattice_group is not None:
            tmp_intensities, _ = _map_space_group_to_input_cell(
                self.intensities, self.params.lattice_group.group()
            )
            self.params.lattice_group = tmp_intensities.space_group_info()

    def _intialise_target(self):
        if self.params.dimensions is Auto:
            dimensions = None
        else:
            dimensions = self.params.dimensions
        if self.params.lattice_group is not None:
            self.lattice_group = (
                self.params.lattice_group.group()
                .build_derived_patterson_group()
                .info()
                .primitive_setting()
                .group()
            )
        self.target = target.Target(
            self.intensities,
            self.dataset_ids.as_numpy_array(),
            min_pairs=self.params.min_pairs,
            lattice_group=self.lattice_group,
            dimensions=dimensions,
            weights=self.params.weights,
        )

    def _determine_dimensions(self):
        if self.params.dimensions is Auto and self.target.dim == 2:
            self.params.dimensions = 2
        elif self.params.dimensions is Auto:
            logger.info("=" * 80)
            logger.info(
                "\nAutomatic determination of number of dimensions for analysis"
            )
            dimensions = []
            functional = []
            for dim in range(1, self.target.dim + 1):
                logger.debug("Testing dimension: %i", dim)
                self.target.set_dimensions(dim)
                max_calls = self.params.minimization.max_calls
                self._optimise(
                    self.params.minimization.engine,
                    max_iterations=self.params.minimization.max_iterations,
                    max_calls=min(20, max_calls) if max_calls else max_calls,
                )
                dimensions.append(dim)
                functional.append(self.minimizer.fun)

            # Find the elbow point of the curve, in the same manner as that used by
            # distl spotfinder for resolution method 1 (Zhang et al 2006).
            # See also dials/algorithms/spot_finding/per_image_analysis.py

            x = np.array(dimensions)
            y = np.array(functional)
            slopes = (y[-1] - y[:-1]) / (x[-1] - x[:-1])
            p_m = slopes.argmin()

            x1 = matrix.col((x[p_m], y[p_m]))
            x2 = matrix.col((x[-1], y[-1]))

            gaps = []
            v = matrix.col(((x2[1] - x1[1]), -(x2[0] - x1[0]))).normalize()

            for i in range(p_m, len(x)):
                x0 = matrix.col((x[i], y[i]))
                r = x1 - x0
                g = abs(v.dot(r))
                gaps.append(g)

            p_g = np.array(gaps).argmax()

            x_g = x[p_g + p_m]

            logger.info(
                dials.util.tabulate(
                    zip(dimensions, functional), headers=("Dimensions", "Functional")
                )
            )
            logger.info("Best number of dimensions: %i", x_g)
            self.target.set_dimensions(int(x_g))
            logger.info("Using %i dimensions for analysis", self.target.dim)

    def run(self):
        self._intialise_target()
        self._determine_dimensions()
        self._optimise(
            self.params.minimization.engine,
            max_iterations=self.params.minimization.max_iterations,
            max_calls=self.params.minimization.max_calls,
        )
        self._principal_component_analysis()

        self._analyse_symmetry()

    @Subject.notify_event(event="optimised")
    def _optimise(self, engine, max_iterations=None, max_calls=None):
        NN = len(set(self.dataset_ids))
        n_sym_ops = len(self.target.sym_ops)

        coords = np.random.rand(NN * n_sym_ops * self.target.dim)
        if engine == "scitbx":
            self.minimizer = cosym_engine.minimize_scitbx_lbfgs(
                self.target,
                coords,
                use_curvatures=self.params.use_curvatures,
                max_iterations=max_iterations,
                max_calls=max_calls,
            )
        else:
            self.minimizer = cosym_engine.minimize_scipy(
                self.target,
                coords,
                method="L-BFGS-B",
                max_iterations=max_iterations,
                max_calls=max_calls,
            )

        self.coords = self.minimizer.x.reshape(
            self.target.dim, NN * n_sym_ops
        ).transpose()

    def _principal_component_analysis(self):
        # Perform PCA
        from sklearn.decomposition import PCA

        pca = PCA().fit(self.coords)
        logger.info("Principal component analysis:")
        logger.info(
            "Explained variance: "
            + ", ".join(["%.2g" % v for v in pca.explained_variance_])
        )
        logger.info(
            "Explained variance ratio: "
            + ", ".join(["%.2g" % v for v in pca.explained_variance_ratio_])
        )
        self.explained_variance = pca.explained_variance_
        self.explained_variance_ratio = pca.explained_variance_ratio_
        if self.target.dim > 3:
            pca.n_components = 3
        self.coords_reduced = pca.fit_transform(self.coords)

    @Subject.notify_event(event="analysed_symmetry")
    def _analyse_symmetry(self):
        sym_ops = [sgtbx.rt_mx(s).new_denominators(1, 12) for s in self.target.sym_ops]

        if not self.input_space_group:
            self._symmetry_analysis = SymmetryAnalysis(
                self.coords, sym_ops, self.subgroups, self.cb_op_inp_min
            )
            logger.info(str(self._symmetry_analysis))
            self.best_solution = self._symmetry_analysis.best_solution
            self.best_subgroup = self.best_solution.subgroup
        else:
            self.best_solution = None
            self._symmetry_analysis = None

        cosets = sgtbx.cosets.left_decomposition(
            self.target._lattice_group,
            self.best_subgroup["subsym"].space_group().build_derived_acentric_group(),
        )
        self.reindexing_ops = self._reindexing_ops(self.coords, sym_ops, cosets)

    def _reindexing_ops(
        self,
        coords: np.ndarray,
        sym_ops: List[sgtbx.rt_mx],
        cosets: sgtbx.cosets.left_decomposition,
    ) -> List[sgtbx.change_of_basis_op]:
        """Identify the reindexing operator for each dataset.

        Args:
          coords (np.ndarray):
            A flattened list of the N-dimensional vectors, i.e. coordinates in
            the first dimension are stored first, followed by the coordinates in
            the second dimension, etc.
          sym_ops (List[sgtbx.rt_mx]): List of cctbx.sgtbx.rt_mx used for the cosym
            symmetry analysis
          cosets (sgtbx.cosets.left_decomposition): The left coset decomposition of the
            lattice group with respect to the proposed Patterson group

        Returns:
          List[sgtbx.change_of_basis_op]: A list of reindexing operators corresponding
            to each dataset.
        """
        reindexing_ops = []
        n_datasets = len(set(self.dataset_ids))
        n_sym_ops = len(sym_ops)
        coord_ids = np.arange(n_datasets * n_sym_ops)
        dataset_ids = coord_ids % n_datasets

        if self.seed_dataset is not None:
            # if seed dataset was specified, use the reindexing op xyz as seed
            sel = np.where(dataset_ids == self.seed_dataset)
            xis = np.array([coords[sel][0]])
        else:
            # choose a high density point as seed
            X = coords
            nbrs = NearestNeighbors(
                n_neighbors=min(11, len(X)), algorithm="brute", metric="cosine"
            ).fit(X)
            distances, indices = nbrs.kneighbors(X)
            average_distance = np.array([dist[1:].mean() for dist in distances])
            i = average_distance.argmin()
            xis = np.array([X[i]])
        coordstr = ",".join(str(round(i, 4)) for i in xis[0])
        logger.debug(f"Coordinate of cluster seed dataset: {coordstr}")

        for j in range(n_datasets):
            sel = np.where(dataset_ids == j)
            X = coords[sel]
            # Find nearest neighbour in cosine-space to the current cluster centroid
            nbrs = NearestNeighbors(
                n_neighbors=min(1, len(X)), algorithm="brute", metric="cosine"
            ).fit(X)
            distances, indices = nbrs.kneighbors([xis.mean(axis=0)])
            k = indices[0][0]
            xis = np.append(xis, [X[k]], axis=0)
            for partition in cosets.partitions:
                if sym_ops[k] in partition:
                    cb_op = sgtbx.change_of_basis_op(partition[0]).new_denominators(
                        self.cb_op_inp_min
                    )
                    reindexing_ops.append(
                        (
                            self.cb_op_inp_min.inverse() * cb_op * self.cb_op_inp_min
                        ).as_xyz()
                    )
                    break

        return reindexing_ops

    def as_dict(self):
        """Return a dictionary representation of the results.

        Returns:
          dict
        """
        d = {
            "input_symmetry": {
                "hall_symbol": self.input_intensities[0]
                .space_group()
                .type()
                .hall_symbol(),
                "unit_cell": self.median_unit_cell.parameters(),
            },
            "cb_op_inp_min": self.cb_op_inp_min.as_xyz(),
            "min_cell_symmetry": {
                "hall_symbol": self.intensities.space_group().type().hall_symbol(),
                "unit_cell": self.intensities.unit_cell().parameters(),
            },
            "lattice_point_group": self.lattice_group.type().hall_symbol(),
        }

        if self._symmetry_analysis is not None:
            d.update(self._symmetry_analysis.as_dict())
        return d

    def as_json(self, filename=None, indent=2):
        """Return a json representation of the results.

        Args:
          filename (str): Optional filename to export the json representation of
            the results.
          indent (int): The indent level for pretty-printing of the json. If ``None``
            is the most compact representation.

        Returns:
          str:
        """
        d = self.as_dict()

        json_str = json.dumps(d, indent=indent)
        if filename:
            with open(filename, "w") as f:
                f.write(json_str)
        return json.dumps(d, indent=indent)


class SymmetryAnalysis:
    def __init__(self, coords, sym_ops, subgroups, cb_op_inp_min):

        import scipy.spatial.distance as ssd

        self.subgroups = subgroups
        self.cb_op_inp_min = cb_op_inp_min
        n_datasets = coords.shape[0] // len(sym_ops)
        dist_mat = ssd.pdist(coords, metric="cosine")
        cos_angle = 1 - ssd.squareform(dist_mat)

        self._sym_ops_cos_angle = {}
        for dataset_id in range(n_datasets):
            for ref_sym_op_id in range(len(sym_ops)):
                ref_idx = n_datasets * ref_sym_op_id + dataset_id
                for sym_op_id in range(ref_sym_op_id + 1, len(sym_ops)):
                    op = sym_ops[ref_sym_op_id].inverse().multiply(sym_ops[sym_op_id])
                    op = op.new_denominators(1, 12)
                    comp_idx = n_datasets * sym_op_id + dataset_id
                    self._sym_ops_cos_angle.setdefault(op, [])
                    self._sym_ops_cos_angle[op].append(cos_angle[ref_idx, comp_idx])

        self._score_symmetry_elements()
        self._score_laue_groups()

    def _score_symmetry_elements(self):
        self.sym_op_scores = {}
        for op, cos_angle in self._sym_ops_cos_angle.items():
            cc_true = 1
            cc = np.mean(cos_angle)
            score = ScoreSymmetryElement(cc, sigma_cc=0.1, cc_true=cc_true)
            score.sym_op = op
            self.sym_op_scores[op] = score

    def _score_laue_groups(self):
        subgroup_scores = [
            ScoreSubGroup(subgrp, list(self.sym_op_scores.values()))
            for subgrp in self.subgroups.result_groups
        ]
        total_likelihood = sum(score.likelihood for score in subgroup_scores)
        for score in subgroup_scores:
            score.likelihood /= total_likelihood
        self.subgroup_scores = sorted(
            subgroup_scores, key=lambda score: score.likelihood, reverse=True
        )

        # The 'confidence' scores are derived from the total probability of the best
        # solution p_best and that for the next best solution p_next:
        #   confidence = [p_best * (p_best - p_next)]^1/2.

        for i, score in enumerate(self.subgroup_scores[:-1]):
            next_score = self.subgroup_scores[i + 1]
            if score.likelihood > 0 and next_score.likelihood > 0:
                lgc = score.likelihood * (score.likelihood - next_score.likelihood)
                confidence = abs(lgc) ** 0.5
                if lgc < 0:
                    confidence = -confidence
                score.confidence = confidence

        self.best_solution = self.subgroup_scores[0]

    @staticmethod
    def sym_ops_table(d):
        header = ("likelihood", "Z-CC", "CC", "", "Operator")
        rows = [header]
        for score in d["sym_op_scores"]:
            rows.append(
                (
                    f"{score['likelihood']:.3f}",
                    f"{score['z_cc']:.2f}",
                    f"{score['cc']:.2f}",
                    score["stars"],
                    str(sgtbx.rt_mx(str(score["operator"])).r().info()),
                )
            )
        return rows

    @staticmethod
    def subgroups_table(d):
        header = (
            "Patterson group",
            "",
            "Likelihood",
            "NetZcc",
            "Zcc+",
            "Zcc-",
            "delta",
            "Reindex operator",
        )
        rows = [header]
        for score in d["subgroup_scores"]:
            rows.append(
                (
                    str(
                        sgtbx.space_group(
                            hall_symbol=str(score["patterson_group"])
                        ).info()
                    ),
                    score["stars"],
                    f"{score['likelihood']:.3f}",
                    f"{score['z_cc_net']: .2f}",
                    f"{score['z_cc_for']: .2f}",
                    f"{score['z_cc_against']: .2f}",
                    f"{score['max_angular_difference']:.1f}",
                    str(sgtbx.change_of_basis_op(str(score["cb_op"]))),
                )
            )
        return rows

    @staticmethod
    def summary_table(d):
        best_subgroup = d["subgroup_scores"][0]
        return (
            (
                "Best solution",
                str(
                    sgtbx.space_group(
                        hall_symbol=str(best_subgroup["patterson_group"])
                    ).info()
                ),
            ),
            (
                "Unit cell",
                "%.3f %.3f %.3f %.1f %.1f %.1f" % tuple(best_subgroup["unit_cell"]),
            ),
            ("Reindex operator", best_subgroup["cb_op"]),
            ("Laue group probability", f"{best_subgroup['likelihood']:.3f}"),
            ("Laue group confidence", f"{best_subgroup['confidence']:.3f}"),
        )

    def __str__(self):
        """Return a string representation of the results.

        Returns:
          str:
        """
        output = []
        output.append("Scoring individual symmetry elements")
        d = self.as_dict()
        output.append(dials.util.tabulate(self.sym_ops_table(d), headers="firstrow"))

        output.append("Scoring all possible sub-groups")
        output.append(dials.util.tabulate(self.subgroups_table(d), headers="firstrow"))

        output.append(
            "Best solution: %s"
            % self.best_solution.subgroup["best_subsym"].space_group_info()
        )
        output.append(
            f"Unit cell: {str(self.best_solution.subgroup['best_subsym'].unit_cell())}"
        )
        output.append(
            "Reindex operator: %s"
            % (self.best_solution.subgroup["cb_op_inp_best"] * self.cb_op_inp_min)
        )
        output.append(f"Laue group probability: {self.best_solution.likelihood:.3f}")
        output.append(f"Laue group confidence: {self.best_solution.confidence:.3f}")
        return "\n".join(output)

    def as_dict(self):
        """Return a dictionary representation of the results.

        Returns:
          dict
        """
        d = {"cb_op_inp_min": self.cb_op_inp_min.as_xyz()}

        d["sym_op_scores"] = []
        for rt_mx, score in self.sym_op_scores.items():
            dd = score.as_dict()
            dd["operator"] = rt_mx.as_xyz()
            d["sym_op_scores"].append(dd)

        d["subgroup_scores"] = []
        for score in self.subgroup_scores:
            dd = score.as_dict()
            dd["cb_op"] = (
                sgtbx.change_of_basis_op(dd["cb_op"]) * self.cb_op_inp_min
            ).as_xyz()
            d["subgroup_scores"].append(dd)

        return d


class ScoreSymmetryElement:
    """Analyse intensities for presence of a given symmetry operation.

    1) Calculate the probability of observing this CC if the sym op is present,
       p(CC; S), modelled by a Cauchy distribution centred on cc_true and width
       gamma = sigma_cc.

    2) Calculate the probability of observing this CC if the sym op is
       NOT present, p(CC; !S).

    3) Calculate the likelihood of symmetry element being present,
       p(S; CC) = p(CC; S) / (p(CC; S) + p(CC; !S))

    See appendix A1 of `Evans, P. R. (2011). Acta Cryst. D67, 282-292.
    <https://doi.org/10.1107/S090744491003982X>`_
    """

    def __init__(self, cc, sigma_cc, cc_true):
        """Initialise a ScoreSymmetryElement object.

        Args:
          cc (float): the correlation coefficient for this symmetry element
          sigma_cc (float): the estimated error in the correlation coefficient
          cc_true (float): the expected value of CC if the symmetry element is present,
            E(CC; S)
        """

        self.cc = cc
        self.sigma_cc = sigma_cc
        self.z_cc = self.cc / self.sigma_cc
        score_cc = ScoreCorrelationCoefficient(self.cc, self.sigma_cc, cc_true)
        self.p_cc_given_s = score_cc.p_cc_given_s
        self.p_cc_given_not_s = score_cc.p_cc_given_not_s
        self.likelihood = score_cc.p_s_given_cc

    @property
    def stars(self):
        # define stars attribute - used mainly for output
        if self.likelihood > 0.9:
            stars = "***"
        elif self.likelihood > 0.7:
            stars = "**"
        elif self.likelihood > 0.5:
            stars = "*"
        else:
            stars = ""
        return stars

    def as_dict(self):
        """Return a dictionary representation of the symmetry element scoring.

        The dictionary will contain the following keys:
          - likelihood: The likelihood of the symmetry element being present
          - z_cc: The Z-score for the correlation coefficient
          - cc: The correlation coefficient for the symmetry element
          - operator: The xyz representation of the symmetry element

        Returns:
          dict:
        """

        return {
            "likelihood": self.likelihood,
            "z_cc": self.z_cc,
            "cc": self.cc,
            "stars": self.stars,
        }


class ScoreSubGroup:
    """Score the probability of a given subgroup being the true subgroup.

    1) Calculates overall Zcc scores for symmetry elements present/absent from
       the subgroup.

    2) Calculates the overall likelihood for this subgroup.

    See appendix A2 of `Evans, P. R. (2011). Acta Cryst. D67, 282-292.
    <https://doi.org/10.1107/S090744491003982X>`_
    """

    def __init__(self, subgroup, sym_op_scores):
        """Initialise a ScoreSubGroup object.

        Args:
          subgroup (dict): A dictionary describing the subgroup as generated by
            :class:`cctbx.sgtbx.lattice_symmetry.metric_subgroups`.
          sym_op_scores (list): A list of :class:`ScoreSymmetryElement` objects for each
            symmetry element possibly in the lattice symmetry.
        """
        # Combined correlation coefficients for symmetry operations
        # present/absent from subgroup
        self.subgroup = subgroup
        patterson_group = subgroup["subsym"].space_group()

        # Overall Zcc scores for symmetry elements present/absent from subgroup
        self.z_cc_for = 0
        self.z_cc_against = 0
        n_for = 0
        n_against = 0
        PL_for = 0
        PL_against = 0
        power = 2
        for score in sym_op_scores:
            if score.sym_op in patterson_group:
                self.z_cc_for += score.z_cc**power
                n_for += 1
                PL_for += math.log(score.p_cc_given_s)
            else:
                self.z_cc_against += score.z_cc**power
                n_against += 1
                PL_against += math.log(score.p_cc_given_not_s)

        # Overall likelihood for this subgroup
        self.likelihood = math.exp(PL_for + PL_against)

        if n_against > 0:
            self.z_cc_against = (self.z_cc_against / n_against) ** (1 / power)
        if n_for > 0:
            self.z_cc_for = (self.z_cc_for / n_for) ** (1 / power)
        self.z_cc_net = self.z_cc_for - self.z_cc_against
        self.confidence = 0

    def __str__(self):
        """Return a string representation of the subgroup scores.

        Returns:
          str:
        """
        return "{} {:.3f} {:.2f} {:.2f} {:.2f}".format(
            self.subgroup["best_subsym"].space_group_info(),
            self.likelihood,
            self.z_cc_net,
            self.z_cc_for,
            self.z_cc_against,
        )

    @property
    def stars(self):
        if self.likelihood > 0.8:
            stars = "***"
        elif self.likelihood > 0.6:
            stars = "**"
        elif self.likelihood > 0.4:
            stars = "*"
        else:
            stars = ""
        return stars

    def as_dict(self):
        """Return a dictionary representation of the subgroup scoring.

        The dictionary will contain the following keys:
          - patterson_group: The current subgroup
          - likelihood: The likelihood of the subgroup being correct
          - confidence: The confidence of the subgroup being correct
          - z_cc_for: The combined Z-scores for all symmetry elements present in the
            subgroup
          - z_cc_against: The combined Z-scores for all symmetry elements present in
            the lattice group but not in the subgroup
          - z_cc_net: The net Z-score, i.e. z_cc_for - z_cc_against
          - max_angular_difference: The maximum angular difference between the
            symmetrised unit cell and the P1 unit cell.
          - cb_op: The change of basis operation from the input unit cell to the
            'best' unit cell.

        Returns:
          dict:
        """
        return {
            "patterson_group": self.subgroup["best_subsym"]
            .space_group()
            .type()
            .hall_symbol(),
            "unit_cell": self.subgroup["best_subsym"].unit_cell().parameters(),
            "likelihood": self.likelihood,
            "confidence": self.confidence,
            "z_cc_net": self.z_cc_net,
            "z_cc_for": self.z_cc_for,
            "z_cc_against": self.z_cc_against,
            "max_angular_difference": self.subgroup["max_angular_difference"],
            "cb_op": f"{self.subgroup['cb_op_inp_best']}",
            "stars": self.stars,
        }


def extract_reference_intensities(params: iotbx.phil.scope_extract) -> miller.array:
    # Extract/calculate a set of intensities from a reference.
    if params.d_min not in {Auto, None}:
        reference_intensities = intensities_from_reference_file(
            params.reference, d_min=params.d_min
        )
    else:
        reference_intensities = intensities_from_reference_file(params.reference)
    initial_space_group_info = reference_intensities.space_group_info()
    group = metric_subgroups(
        reference_intensities.crystal_symmetry(),
        params.lattice_symmetry_max_delta,
        enforce_max_delta_for_generated_two_folds=True,
    ).result_groups[0]
    ref_cb_op = (
        group["best_subsym"].change_of_basis_op_to_minimum_cell()
        * group["cb_op_inp_best"]
    )

    reference_intensities = (
        reference_intensities.change_basis(
            ref_cb_op,
        )
        .expand_to_p1()
        .as_non_anomalous_array()
        .merge_equivalents()
        .array()
    )
    if not reference_intensities.sigmas():
        reference_intensities.set_sigmas(flex.double(reference_intensities.size(), 1))
    return reference_intensities, initial_space_group_info
